/**
 * @file interface.c
 *
 * @brief Service Placement interface implementation for Linux.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: interface.c 2548 2008-02-13 19:20:58Z wittenbu $
 */

#include <errno.h>
#include <error.h>
#include <netdb.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>

#include "../../types.h"
#include "../../vars.h"
#include "../../interface.h"
#include "../../util/debug.h"
#include "../../util/net.h"



/**
 * @brief Linux specific implemenation of __SP_vprintf().
 */
void __SP_vprintf(const char *format, va_list ap) {
	vprintf(format, ap);
}

/**
 * @brief Linux specific implemenation of __SP_random().
 */
double __SP_random(double max) {
	return (max * ((double) random() / (double) RAND_MAX));
}

/**
 * @brief Linux specific implemenation of __SP_time().
 */
SP_time_t __SP_time() {
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return ((SP_time_t) tv.tv_sec) + (((SP_time_t) tv.tv_usec) / 1000000);
}

/**
 * @brief Linux specific implemenation of __SP_send().
 */
void __SP_send(const SP_packet_header_t* header, const SP_packet_payload_t* payload) {
	int packet_size = sizeof(SP_packet_header_t) + header->payload_length;

	SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Sending packet of %d byte(s) to node %s (seqno=%d)...\n", packet_size, addr_to_str(header->to), header->seqno);

	// Copy packet header and payload to buffer.
	char *buffer;
	SP_MALLOC_ARRAY(buffer, char, packet_size);
	memcpy(buffer, header, sizeof(SP_packet_header_t));
	if(header->payload_length > 0)
		memcpy(buffer + sizeof(SP_packet_header_t), payload, header->payload_length);

	// Assemable address information.
	struct sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = header->to;
	addr.sin_port = htons(SP_DEFAULT_UDP_PORT);

	// Send packet.
	if(sendto(SP_udp_socket, buffer, packet_size, 0, (struct sockaddr*) &addr, sizeof(struct sockaddr_in)) < 0) {
		int errsv = errno;
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "Sending packet failed. (%d)\n", errsv);
		error(0, errsv, NULL);
	}
}

/**
 * @brief Linux specific implemenation of SP_active_callback().
 *
 * Callback activation is handled in the main loop, hence there is nothing to
 * do.
 */
void __SP_activate_callback(SP_time_t time) {
	// Nothing to do on Linux.
}
