/**
 * @file sp_application.cc
 *
 * @brief Service Placement application for ns-2.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: sp_application.cc 2548 2008-02-13 19:20:58Z wittenbu $
 */

#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <queue>
#include <vector>

#include "app.h"
#include "packet.h"
#include "rng.h"
#include "scheduler.h"
#include "timer-handler.h"

#include "sp_application.h"



/**
 * @brief Group the output of vprintf() by time.
 */
#define VPRINTF_GROUP_OUTPUT_BY_TIME	0



#define STATE_INIT	0x0	///< Application is initializing.
#define STATE_STOPPED	0x1	///< Application is stopped.
#define STATE_RUNNING	0x2	///< Application is running.
#define STATE_TRANSMIT	0x3	///< Application is transmitting a packet.
#define STATE_FINALIZED	0x4	///< Application is finalized.

/**
 * @brief Convert a numeric state represenstation into the corresponding
 * string.
 *
 * @param state Numeric representation of the state.
 * @return String representing the state.
 */
const char* STATE_TO_STRING(int state) {
	switch(state) {
	case STATE_INIT:	return "INIT";
	case STATE_STOPPED:	return "STOPPED";
	case STATE_RUNNING:	return "RUNNING";
	case STATE_TRANSMIT:	return "TRANSMIT";
	case STATE_FINALIZED:	return "FINALIZED";
	default: printf("\t\t\tServicePlacementApplication: Illegal state (%d), aborting...\n", state); abort();
	}
}



using namespace ServicePlacement;

// Glue code to the Tcl interpreter.
static class ServicePlacementApplicationClass : public TclClass {
public:
	ServicePlacementApplicationClass() : TclClass("Application/ServicePlacement") {}
	TclObject* create(int, const char*const*) {
		return (new ServicePlacementApplication());
	}
} class_service_placement_application;



ServicePlacementApplication* ServicePlacementApplication::instance_ = NULL;
double ServicePlacementApplication::timeOfLastCallToVprintf_ = 0;

/**
 * @brief ServicePlacementApplication constructor.
 */
ServicePlacementApplication::ServicePlacementApplication() : \
	Application()
{
	if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_VERBOSE)
		printf("\t\t\tServicePlacementApplication: Creating new instance...\n");
	instance_ = this;
	timer_ = new ServicePlacementApplicationTimer(this);
	rng_ = new RNG();
	rng_->set_seed(RNG::PREDEF_SEED_SOURCE, 0);
	state_ = STATE_INIT;
	timeOfNextCallback_ = 0;

	// Initialize and export statistics variables.
	transmittedUnicastPackets_ = 0;
	bind("transmittedUnicastPackets_", &transmittedUnicastPackets_);
	receivedUnicastPackets_ = 0;
	bind("receivedUnicastPackets_", &receivedUnicastPackets_);
	transmittedBroadcastPackets_ = 0;
	bind("transmittedBroadcastPackets_", &transmittedBroadcastPackets_);
	receivedBroadcastPackets_ = 0;
	bind("receivedBroadcastPackets_", &receivedBroadcastPackets_);
	transmittedBytes_ = 0;
	bind("transmittedBytes_", &transmittedBytes_);
	receivedBytes_ = 0;
	bind("receivedBytes_", &receivedBytes_);

	// Export C variables.
	bind("SP_direct_neighbors", &SP_direct_neighbors);
	bind("SP_longest_route", &SP_longest_route);
	bind("SP_average_route", &SP_average_route);
	bind("SP_DEBUG_ERRORs", &SP_DEBUG_ERRORs);
	bind("SP_DEBUG_WARNs", &SP_DEBUG_WARNs);
	bind("SP_RUDP_total_transmissions", &SP_RUDP_total_transmissions);
	bind("SP_RUDP_successful_transmissions", &SP_RUDP_successful_transmissions);
	bind("SP_ICMP_transmitted_nops", &SP_ICMP_transmitted_nops);
	bind("SP_ICMP_received_nops", &SP_ICMP_received_nops);
	bind("SP_SLP_total_queries", &SP_SLP_total_queries);
	bind("SP_SLP_successful_queries", &SP_SLP_successful_queries);
	bind("SP_SERVICE_total_data_requests", &SP_SERVICE_total_data_requests);
	bind("SP_SERVICE_successful_data_requests", &SP_SERVICE_successful_data_requests);
	bind("SP_SERVICE_total_successful_data_request_delay", &SP_SERVICE_total_successful_data_request_delay);
	bind("SP_SERVICE_total_data_updates", &SP_SERVICE_total_data_updates);
	bind("SP_SERVICE_successful_data_updates", &SP_SERVICE_successful_data_updates);
	bind("SP_SERVICE_total_stream_requests", &SP_SERVICE_total_stream_requests);
	bind("SP_SERVICE_successful_stream_requests", &SP_SERVICE_successful_stream_requests);
	bind("SP_SERVICE_expected_stream_packets", &SP_SERVICE_expected_stream_packets);
	bind("SP_SERVICE_received_stream_packets", &SP_SERVICE_received_stream_packets);
	bind("SP_SERVICE_total_replications", &SP_SERVICE_total_replications);
	bind("SP_SERVICE_successful_replications", &SP_SERVICE_successful_replications);
	bind("SP_SERVICE_total_migrations", &SP_SERVICE_total_migrations);
	bind("SP_SERVICE_successful_migrations", &SP_SERVICE_successful_migrations);

	// Call C code initializer.
	state_ = STATE_STOPPED;
	SP_init();
}

/**
 * @brief ServicePlacementApplication destructor.
 */
ServicePlacementApplication::~ServicePlacementApplication() {
	if(state_ != STATE_FINALIZED) {
		if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
			printf("\t\t\tServicePlacementApplication: Trying to delete application, but it is not finalized (%s)...\n", STATE_TO_STRING(state_));
		abort();
	}

	if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_VERBOSE)
		printf("\t\t\tServicePlacementApplication: Deleting application...\n");

	// Cancel any remaining callbacks.
	if(timeOfNextCallback_ != 0)
		timer_->cancel();

	delete timer_;
	delete rng_;
}

/*
 * ns-2 entry points.
 */

int ServicePlacementApplication::command(int argc, const char*const* argv) {
	if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_VERBOSE)
		printf("\t\t\tServicePlacementApplication: Processing command \"%s\"...\n", argv[1]);
	instance_ = this;

//	Tcl& tcl = Tcl::instance();
	if(argc == 2) {
		if(strcmp(argv[1], "finalize") == 0) {
			SP_finalize();
			state_ = STATE_FINALIZED;
			fflush(stdout);	// Flush standard output streams.
			return TCL_OK;
		}
	} else if(argc == 3) {
		// Intercept "attach-agent" command and initialize the local address.
		if(strcmp(argv[1], "attach-agent") == 0) {
			int ret = Application::command(argc, argv);
			if(ret == TCL_OK)
				SP_local_addr = agent_->addr();
			return ret;
		}
	} else if(argc == 4) {
		if(strcmp(argv[1], "seed") == 0) {
			int i = atoi(argv[3]);
			if(strcmp(argv[2], "raw") == 0) {
				rng_->set_seed(RNG::RAW_SEED_SOURCE, i);
			} else if(strcmp(argv[2], "predef") == 0) {
				rng_->set_seed(RNG::PREDEF_SEED_SOURCE, i);
			} else if(strcmp(argv[2], "heuristic") == 0) {
				rng_->set_seed(RNG::HEURISTIC_SEED_SOURCE, 0);
			}
			return TCL_OK;
		}
	}

	if(SP_command(argc, argv))
		return TCL_OK;

	return Application::command(argc, argv);
}

void ServicePlacementApplication::start() {
	if(state_ != STATE_STOPPED) {
		if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
			printf("\t\t\tServicePlacementApplication: Trying to start application, but it is not stopped (%s)...\n", STATE_TO_STRING(state_));
		abort();
	}

	if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_VERBOSE)
		printf("\t\t\tServicePlacementApplication: Starting...\n");
	instance_ = this;
	state_ = STATE_RUNNING;
	SP_start();
}

void ServicePlacementApplication::stop() {
	if(state_ != STATE_RUNNING) {
		if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
			printf("\t\t\tServicePlacementApplication: Trying to stop application, but it is not running (%s)...\n", STATE_TO_STRING(state_));
		abort();
	}

	if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_VERBOSE)
		printf("\t\t\tServicePlacementApplication: Stopping...\n");
	instance_ = this;
	SP_stop();
	state_ = STATE_STOPPED;
}

// Unused ns-2 entry point.
void ServicePlacementApplication::send(int nbytes) {
	if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
		printf("\t\t\tServicePlacementApplication: Sending %d byte(s) on unused interface.\n", nbytes);
	abort();
}

// Unused ns-2 entry point.
void ServicePlacementApplication::recv(int nbytes) {
	if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
		printf("\t\t\tServicePlacementApplication: Received %d byte(s) on unused interface.\n", nbytes);
	abort();
}

// ns-2 entry point for received packets.
void ServicePlacementApplication::recv(Packet* pkt, Handler*) {
	if(pkt == NULL) {
		if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
			printf("\t\t\tServicePlacementApplication: Received a NULL packet.\n");
		abort();
	} else if(pkt->userdata()->size() < ((int) sizeof(SP_packet_header_t))) {
		if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
			printf("\t\t\tServicePlacementApplication: Received %d byte(s) but expected at least %d bytes.\n", pkt->userdata()->size(), sizeof(SP_packet_header_t));
		abort();
	}

	unsigned char* p = ((PacketData*) pkt->userdata())->data();
	SP_packet_header_t* header = (SP_packet_header_t*) p;
	header->hops += hdr_cmn::access(pkt)->num_forwards();
	header->last_hop = hdr_cmn::access(pkt)->prev_hop();

	SP_packet_payload_t* payload = NULL;
	if(header->payload_length > 0)
		payload = p + sizeof(SP_packet_header_t);

	if((state_ != STATE_RUNNING) && (state_ != STATE_STOPPED)) {
		if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
			printf("\t\t\tServicePlacementApplication: Trying to process %d byte(s) of application data from node #%d, but application is neither running nor stopped...\n", pkt->userdata()->size(), header->from);
		abort();
	}

	if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_VERBOSE)
		printf("\t\t\tServicePlacementApplication: Processing %d byte(s) of application data from node #%d (seqno=%d, hops=%d, last_hop=%d)...\n", pkt->userdata()->size(), header->from, header->seqno, header->hops, header->last_hop);
	instance_ = this;
	SP_recv(header, payload);

	// Statistics.
	if(header->to == SP_BROADCAST_ADDR)
		receivedBroadcastPackets_++;
	else
		receivedUnicastPackets_++;
	receivedBytes_ += pkt->userdata()->size();
}

// Unused ns-2 entry point.
void ServicePlacementApplication::process_data(int size, AppData* data) {
	if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
		printf("\t\t\tServicePlacementApplication: Received %d byte(s) of application data at %p on unused interface.\n\t\t\tServicePlacementApplication: Is processPackets_ in Agent/MessagePassing set to 1?\n", size, data);
	abort();
}

// ns-2 entry point upon completed packet transmission.
void ServicePlacementApplication::resume() {
	if(state_ != STATE_TRANSMIT) {
		if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
			printf("\t\t\tServicePlacementApplication: Trying to resume application, but it is not transmitting (%s)...\n", STATE_TO_STRING(state_));
		abort();
	}

	if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_VERBOSE)
		printf("\t\t\tServicePlacementApplication: Resuming...\n");
	state_ = STATE_RUNNING;
}

/*
 * C code entry points.
 */

/**
 * @brief ns-2 specific implementation of __SP_vprintf().
 */
void ServicePlacementApplication::__SP_vprintf(const char *format, va_list ap) {
	if(VPRINTF_GROUP_OUTPUT_BY_TIME && (timeOfLastCallToVprintf_ != Scheduler::instance().clock()))
		printf("\n");
	vprintf(format, ap);
	timeOfLastCallToVprintf_ = Scheduler::instance().clock();
}

/**
 * @brief ns-2 specific implementation of __SP_random().
 */
double ServicePlacementApplication::__SP_random(double max) {
	if((state_ != STATE_RUNNING) && (state_ != STATE_STOPPED)) {
		if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
			printf("\t\t\tServicePlacementApplication: Trying to generate random number between 0 and %f, but application is neither running nor stopped (%s)...\n", max, STATE_TO_STRING(state_));
		abort();
	}

	if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_VERBOSE)
		printf("\t\t\tServicePlacementApplication: Generating random number between 0 and %f...\n", max);
	return rng_->uniform(max);
}

/**
 * @brief ns-2 specific implementation of __SP_time().
 */
SP_time_t ServicePlacementApplication::__SP_time() {
	return Scheduler::instance().clock();
}

/**
 * @brief ns-2 specific implementation of __SP_send().
 */
void ServicePlacementApplication::__SP_send(const SP_packet_header_t* header, const SP_packet_payload_t* payload) {
	int packet_size = sizeof(SP_packet_header_t) + header->payload_length;

	if((state_ != STATE_RUNNING) && (state_ != STATE_STOPPED)) {
		if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
			printf("\t\t\tServicePlacementApplication: Trying to send packet of %d byte(s) with sequence number %d to node #%d, but application is neither running nor stopped (%s)...\n", packet_size, header->seqno, header->to, STATE_TO_STRING(state_));
		abort();
	}

	if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_VERBOSE)
		printf("\t\t\tServicePlacementApplication: Sending packet of %d byte(s) to node #%d (seqno=%d)...\n", packet_size, header->to, header->seqno);
	state_ = STATE_TRANSMIT;

	ns_addr_t dst = {header->to, agent_->port()};
	PacketData* data = new PacketData(packet_size);
	memcpy(data->data(), header, sizeof(SP_packet_header_t));
	if(header->payload_length > 0)
		memcpy(data->data() + sizeof(SP_packet_header_t), payload, header->payload_length);
	agent_->sendto(packet_size, data, NULL, dst);

	// Statistics.
	if(header->to == SP_BROADCAST_ADDR)
		transmittedBroadcastPackets_++;
	else
		transmittedUnicastPackets_++;
	transmittedBytes_ += packet_size;

}

/**
 * @brief ns-2 specific implementation of __SP_activate_callback().
 */
void ServicePlacementApplication::__SP_activate_callback(SP_time_t time) {
	if((state_ != STATE_RUNNING) && (state_ != STATE_STOPPED)) {
		if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
			printf("\t\t\tServicePlacementApplication: Trying activate callback to be called at %.2fs, but application is neither running nur stopped (%s)...\n", time, STATE_TO_STRING(state_));
		abort();
	}

	if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_VERBOSE)
		printf("\t\t\tServicePlacementApplication: Activating callback to be called at %.2fs...\n", time);

	// Schedule next callback.
	double delay = time - Scheduler::instance().clock();
	if(delay < 0) {
		if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
			printf("\t\t\tServicePlacementApplication: Trying to activate callback in the past...\n");
		abort();
	}
	timeOfNextCallback_ = time;
	timer_->resched(delay);
}

/*
 * Timer entry points.
 */

void ServicePlacementApplication::expire() {
	if((state_ != STATE_RUNNING) && (state_ != STATE_STOPPED)) {
		if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_WARN)
			printf("\t\t\tServicePlacementApplication: Trying execute callback, but application is neither running nor stopped (%s)...\n", STATE_TO_STRING(state_));
		return;
	} else if(timeOfNextCallback_ == 0) {
		if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
			printf("\t\t\tServicePlacementApplication: Timer expired but none was expected.\n");
		abort();
	} else if(Scheduler::instance().clock() != timeOfNextCallback_) {
		if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_FATAL)
			printf("\t\t\tServicePlacementApplication: Timer expired at incorrect time (%f != %f).\n", Scheduler::instance().clock(), timeOfNextCallback_);
		abort();
	}

	if(SP_DEBUG_LEVEL >= SP_DEBUG_LEVEL_VERBOSE)
		printf("\t\t\tServicePlacementApplication: Timer expired, executing callback...\n");

	// Execute callback.
	instance_ = this;
	timeOfNextCallback_ = 0;
	SP_execute_callback();
}



/**
 * ServicePlacementApplicationTimer
 **/

ServicePlacementApplicationTimer::ServicePlacementApplicationTimer(ServicePlacementApplication* app) : \
	TimerHandler(), app_(app)
{}

void ServicePlacementApplicationTimer::expire(Event*) {
	app_->expire();
}



/**
 * Include C code
 **/

// Calls into C code.
#define SP_init ServicePlacementApplication::SP_init
#define SP_start ServicePlacementApplication::SP_start
#define SP_stop ServicePlacementApplication::SP_stop
#define SP_finalize ServicePlacementApplication::SP_finalize
#define SP_recv ServicePlacementApplication::SP_recv
#define SP_execute_callback ServicePlacementApplication::SP_execute_callback
#define SP_command ServicePlacementApplication::SP_command

// Calls from C code.
#define __SP_vprintf ServicePlacementApplication::instance_->__SP_vprintf
#define __SP_random ServicePlacementApplication::instance_->__SP_random
#define __SP_time ServicePlacementApplication::instance_->__SP_time
#define __SP_send ServicePlacementApplication::instance_->__SP_send
#define __SP_activate_callback ServicePlacementApplication::instance_->__SP_activate_callback

// Global C variables.
#define SP_callbacks ServicePlacementApplication::instance_->SP_callbacks
#define SP_callback_time ServicePlacementApplication::instance_->SP_callback_time
#define SP_registered_callbacks ServicePlacementApplication::instance_->SP_registered_callbacks
#define SP_executed_callbacks ServicePlacementApplication::instance_->SP_executed_callbacks
#define SP_local_addr ServicePlacementApplication::instance_->SP_local_addr
#define SP_sequence_number ServicePlacementApplication::instance_->SP_sequence_number
#define SP_received_packet_cache ServicePlacementApplication::instance_->SP_received_packet_cache
#define SP_network_map ServicePlacementApplication::instance_->SP_network_map
#define SP_direct_neighbors ServicePlacementApplication::instance_->SP_direct_neighbors
#define SP_longest_route ServicePlacementApplication::instance_->SP_longest_route
#define SP_average_route ServicePlacementApplication::instance_->SP_average_route
#define SP_addr_to_str_buffer ServicePlacementApplication::instance_->SP_addr_to_str_buffer
#define SP_addr_to_str_buffer_index ServicePlacementApplication::instance_->SP_addr_to_str_buffer_index
#define SP_DEBUG_ERRORs ServicePlacementApplication::instance_->SP_DEBUG_ERRORs
#define SP_DEBUG_WARNs ServicePlacementApplication::instance_->SP_DEBUG_WARNs
// ICMP-light
#define SP_ICMP_active_echo_request ServicePlacementApplication::instance_->SP_ICMP_active_echo_request
#define SP_ICMP_transmitted_nops ServicePlacementApplication::instance_->SP_ICMP_transmitted_nops
#define SP_ICMP_received_nops ServicePlacementApplication::instance_->SP_ICMP_received_nops
// RUDP-light
#define SP_RUDP_transmission_queue ServicePlacementApplication::instance_->SP_RUDP_transmission_queue
#define SP_RUDP_total_transmissions ServicePlacementApplication::instance_->SP_RUDP_total_transmissions
#define SP_RUDP_successful_transmissions ServicePlacementApplication::instance_->SP_RUDP_successful_transmissions
#define SP_RUDP_failed_transmissions ServicePlacementApplication::instance_->SP_RUDP_failed_transmissions
// SLP-light
#define SP_SLP_active_queries ServicePlacementApplication::instance_->SP_SLP_active_queries
#define SP_SLP_total_queries ServicePlacementApplication::instance_->SP_SLP_total_queries
#define SP_SLP_successful_queries ServicePlacementApplication::instance_->SP_SLP_successful_queries
#define SP_SLP_failed_queries ServicePlacementApplication::instance_->SP_SLP_failed_queries
// Service
#define SP_SERVICE_service_cache ServicePlacementApplication::instance_->SP_SERVICE_service_cache
#define SP_SERVICE_total_data_requests ServicePlacementApplication::instance_->SP_SERVICE_total_data_requests
#define SP_SERVICE_successful_data_requests ServicePlacementApplication::instance_->SP_SERVICE_successful_data_requests
#define SP_SERVICE_failed_data_requests ServicePlacementApplication::instance_->SP_SERVICE_failed_data_requests
#define SP_SERVICE_total_successful_data_request_delay ServicePlacementApplication::instance_->SP_SERVICE_total_successful_data_request_delay
#define SP_SERVICE_total_data_updates ServicePlacementApplication::instance_->SP_SERVICE_total_data_updates
#define SP_SERVICE_successful_data_updates ServicePlacementApplication::instance_->SP_SERVICE_successful_data_updates
#define SP_SERVICE_failed_data_updates ServicePlacementApplication::instance_->SP_SERVICE_failed_data_updates
#define SP_SERVICE_total_stream_requests ServicePlacementApplication::instance_->SP_SERVICE_total_stream_requests
#define SP_SERVICE_successful_stream_requests ServicePlacementApplication::instance_->SP_SERVICE_successful_stream_requests
#define SP_SERVICE_failed_stream_requests ServicePlacementApplication::instance_->SP_SERVICE_failed_stream_requests
#define SP_SERVICE_expected_stream_packets ServicePlacementApplication::instance_->SP_SERVICE_expected_stream_packets
#define SP_SERVICE_received_stream_packets ServicePlacementApplication::instance_->SP_SERVICE_received_stream_packets
#define SP_SERVICE_hosted_services ServicePlacementApplication::instance_->SP_SERVICE_hosted_services
#define SP_SERVICE_served_streams ServicePlacementApplication::instance_->SP_SERVICE_served_streams
#define SP_SERVICE_incoming_replications ServicePlacementApplication::instance_->SP_SERVICE_incoming_replications
#define SP_SERVICE_outgoing_replications ServicePlacementApplication::instance_->SP_SERVICE_outgoing_replications
#define SP_SERVICE_total_replications ServicePlacementApplication::instance_->SP_SERVICE_total_replications
#define SP_SERVICE_successful_replications ServicePlacementApplication::instance_->SP_SERVICE_successful_replications
#define SP_SERVICE_failed_replications ServicePlacementApplication::instance_->SP_SERVICE_failed_replications
#define SP_SERVICE_total_migrations ServicePlacementApplication::instance_->SP_SERVICE_total_migrations
#define SP_SERVICE_successful_migrations ServicePlacementApplication::instance_->SP_SERVICE_successful_migrations
#define SP_SERVICE_failed_migrations ServicePlacementApplication::instance_->SP_SERVICE_failed_migrations

// Placement
#define SP_PLACEMENT_active_algorithm ServicePlacementApplication::instance_->SP_PLACEMENT_active_algorithm

// C code files.
#include "../../interface.c"
#include "../../util/debug.c"
#include "../../util/list.c"
#include "../../util/callback.c"
#include "../../util/net.c"
#include "../../icmp-light/icmp-light.c"
#include "../../rudp-light/rudp-light.c"
#include "../../slp-light/slp-light.c"
#include "../../service/service.c"
#include "../../service/service_client.c"
#include "../../service/service_server.c"
#include "../../service/service_management.c"
#include "../../service/service_replication.c"
#include "../../service/service_util.c"
#include "../../placement/placement.c"
#include "../../placement/placement_stats.c"
#include "../../placement/simple_heuristic/simple_heuristic.c"
#include "../../placement/tree-topology_migration/tree-topology_migration.c"
#include "../../placement/linkpull_migration/linkpull_migration.c"
#include "../../placement/peerpull_migration/peerpull_migration.c"
