/**
 * @file icmp-light.c
 *
 * @brief Internet Control Message Protocol (ICMP) - minimal implemenation.
 *
 * A very simple implementation of basic functionality of the Internet Control
 * Message Protocol (ICMP). Supports ECHO and REPLY packets as well as NOP
 * packets that do nothing but causing traffic.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: icmp-light.c 2548 2008-02-13 19:20:58Z wittenbu $
 */

#include <stdlib.h>
#include <string.h>

#include "../types.h"
#include "../vars.h"
#include "../interface.h"
#include "../interface_internal.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "icmp-light.h"
#include "icmp-light_types.h"
#include "icmp-light_vars.h"

#define SP_TYPE_ICMP_NOP		0x1	///< ICMP NOP packet type.
#define SP_TYPE_ICMP_ECHO_REQUEST	0x2	///< ICMP ECHO REQUEST packet type.
#define SP_TYPE_ICMP_ECHO_REPLY		0x3	///< ICMP ECHO REPLY packet type.

/**
 * @brief Maximal supported number of replies to an ECHO REQUEST.
 */
#define ICMP_MAX_ECHO_REQUEST_NODES	1024



/*
 * Interface to main module.
 */

/**
 * @brief Initiliaze the ICMP module.
 */
void ICMP_init() {
	memset(&SP_ICMP_active_echo_request, 0, sizeof(SP_ICMP_echo_request_t));
	SP_ICMP_transmitted_nops = 0;
	SP_ICMP_received_nops = 0;
}

/**
 * @brief Start the ICMP module.
 */
void ICMP_start() {
}

/**
 * @brief Stop the ICMP module.
 */
void ICMP_stop() {
}

/**
 * @brief Finalize the ICMP module.
 */
void ICMP_finalize() {
}



/*
 * Packet handling functions.
 */

/**
 * @brief Send an ICMP NOP packet.
 *
 * @param to Address of the destination node.
 * @param ttl Initial time-to-life of the packet.
 */
static void send_ICMP_NOP(SP_addr_t to, SP_ttl_t ttl) {
	SP_packet_header_t header;

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "ICMP: Sending NOP to node #%d with a TTL of %d...\n", to, ttl);
	init_header(&header);
	header.to = to;
	header.ttl = ttl;
	header.app = SP_APP_ICMP;
	header.type = SP_TYPE_ICMP_NOP;
	SP_send(&header, NULL);
}

/**
 * @brief Send an ICMP ECHO REQUEST packet.
 *
 * @param to Address of the destination node.
 * @param ttl Initial time-to-live of the packet.
 */
static void send_ICMP_ECHO_REQUEST(SP_addr_t to, SP_ttl_t ttl) {
	SP_packet_header_t header;

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "ICMP: Sending ECHO REQUEST to node #%d with a TTL of %d...\n", to, ttl);
	init_header(&header);
	header.to = to;
	header.ttl = ttl;
	header.app = SP_APP_ICMP;
	header.type = SP_TYPE_ICMP_ECHO_REQUEST;
	SP_send(&header, NULL);
}

/**
 * @brief Send an ICMP ECHO REPLY packet.
 *
 * @param to Address of the destination node.
 */
static void send_ICMP_ECHO_REPLY(SP_addr_t to) {
	SP_packet_header_t header;

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "ICMP: Sending ECHO REPLY to node #%d...\n", to);
	init_header(&header);
	header.to = to;
	header.app = SP_APP_ICMP;
	header.type = SP_TYPE_ICMP_ECHO_REPLY;
	SP_send(&header, NULL);
}

/**
 * @brief Handle an incoming ICMP NOP packet.
 *
 * @param header Pointer to header of the incoming packet.
 */
static void handle_SP_TYPE_ICMP_NOP(const SP_packet_header_t* header) {
	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "ICMP: Received NOP from node #%d.\n", header->from);
	if((header->to == SP_BROADCAST_ADDR) && (header->ttl > 1)) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "ICMP: Forwarding NOP...\n");
		forward(header, NULL);
	}

	SP_ICMP_received_nops++;
}

/**
 * @brief Handle an incoming ICMP ECHO REQUEST packet.
 *
 * @param header Pointer to header of the incoming packet.
 */
static void handle_ICMP_ECHO_REQUEST(const SP_packet_header_t* header) {
	if(header->from == SP_LOCAL_ADDR) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "ICMP: Dropping own ECHO REQUEST...\n");
		return;
	}
	SP_DEBUG(SP_DEBUG_LEVEL_INFO, "ICMP: Received ECHO REQUEST from node #%d.\n", header->from);
	send_ICMP_ECHO_REPLY(header->from);
	if((header->to == SP_BROADCAST_ADDR) && (header->ttl > 1)) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "ICMP: Forwarding ECHO REQUEST...\n");
		forward(header, NULL);
	}
}

/**
 * @brief Handle an incoming ICMP ECHO REPLY packet.
 *
 * @param header Pointer to header of the incoming packet.
 */
static void handle_ICMP_ECHO_REPLY(const SP_packet_header_t* header) {
	if(SP_ICMP_active_echo_request.active) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "ICMP: Received ECHO REPLY from node #%d.\n", header->from);
	} else {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "ICMP: Received ECHO REPLY from node #%d, but no echo request in progress.\n", header->from);
		return;
	}
	if(list_size(SP_ICMP_active_echo_request.echo_request_result->nodes) >= SP_ICMP_active_echo_request.max_nodes) {
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "ICMP: Too many replies, dropping ECHO REPLY from node #%d...\n", header->from);
		SP_ICMP_active_echo_request.too_many_replies = true;
	}

	SP_node_t* node;
	SP_MALLOC(node, SP_node_t);
	node->address = header->from;
	node->hops = header->hops;
	list_add(SP_ICMP_active_echo_request.echo_request_result->nodes, node);
}

/**
 * @brief ICMP packet classifier.
 *
 * @param header Pointer to header of an incoming packet.
 * @param payload Pointer to payload of an incoming packet.
 */
void ICMP_recv(const SP_packet_header_t* header, const SP_packet_payload_t* payload) {
	if(header->app != SP_APP_ICMP) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "ICMP: Received non-ICMP packet.\n");
		abort();
	}

	switch(header->type) {
	case SP_TYPE_NONE:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "ICMP: Received packet with uninitialized type field from node #%d.\n", header->from);
		break;
	case SP_TYPE_ICMP_NOP:
		handle_SP_TYPE_ICMP_NOP(header);
		break;
	case SP_TYPE_ICMP_ECHO_REQUEST:
		handle_ICMP_ECHO_REQUEST(header);
		break;
	case SP_TYPE_ICMP_ECHO_REPLY:
		handle_ICMP_ECHO_REPLY(header);
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "ICMP: Received packet of unknown type 0x%X from node %s.\n", header->type, addr_to_str(header->from));
		abort();
	}
}



/*
 * Interface provided by this module.
 */

/**
 * @brief Send an ICMP NOP packet.
 *
 * @param to Address of the destination node.
 * @param ttl Initial time-to-life of the packet.
 */
void ICMP_send_nop(SP_addr_t to, SP_ttl_t ttl) {
	SP_DEBUG(SP_DEBUG_LEVEL_NOTE, "ICMP: Sending NOP to node %s with a TTL of %d...\n", addr_to_str(to), ttl);
	send_ICMP_NOP(to, ttl);

	SP_ICMP_transmitted_nops++;
}

/**
 * @brief Timeout callback for ICMP ECHO REQUESTs.
 *
 * @param parameter Echo request data structure.
 */
void ICMP_send_echo_request_timeout(SP_callback_parameter_t parameter) {
	SP_ICMP_echo_request_t* echo_request = (SP_ICMP_echo_request_t*) parameter;

	if(!list_empty(echo_request->echo_request_result->nodes)) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "ICMP: Timeout for ECHO REQUEST has expired, calling callback function...\n");
		if(echo_request->too_many_replies)
			echo_request->echo_request_result->errno = SP_ERROR_OUT_OF_MEMORY;
		else
			echo_request->echo_request_result->errno = SP_ERROR_OK;
	} else {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "ICMP: Timeout for ECHO REQUEST has expired without any replies.\n");
		echo_request->echo_request_result->errno = SP_ERROR_TIMEOUT;
	}
	echo_request->function(echo_request->echo_request_result);
	memset(echo_request, 0, sizeof(SP_ICMP_echo_request_t));
}

/**
 * @brief Send an ICMP ECHO REQUEST packet.
 *
 * @param to Address of the destination node.
 * @param ttl Initial time-to-life of the packet.
 * @param max_nodes Maximal number of replies to report.
 * @param function Pointer to function to call upon completion.
 * @param echo_request_result Data structure for the requests of the echo request.
 */
void ICMP_send_echo_request(SP_addr_t to, SP_ttl_t ttl, unsigned int max_nodes, SP_callback_function_t function, SP_ICMP_echo_request_result_t* echo_request_result) {
	if((function == NULL) || (echo_request_result == NULL)) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SLP: Callback function or parameter not given (%p, %p).\n", function, echo_request_result);
		return;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_NOTE, "ICMP: Sending echo request to node %s with a TTL of %d...\n", addr_to_str(to), ttl);
	if(SP_ICMP_active_echo_request.active) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "ICMP: Echo request in progress.\n");
		echo_request_result->errno = SP_ERROR_BUSY;
		function(echo_request_result);
		return;
	}

	SP_ICMP_active_echo_request.active = true;
	SP_ICMP_active_echo_request.max_nodes = max_nodes;
	SP_ICMP_active_echo_request.function = function;
	SP_ICMP_active_echo_request.echo_request_result = echo_request_result;
	SP_ICMP_active_echo_request.too_many_replies = false;

	send_ICMP_ECHO_REQUEST(to, ttl);
	callback_add(ICMP_send_echo_request_timeout, &SP_ICMP_active_echo_request, SP_TIMEOUT);
}



/*
 * Interface provided by this module for direct commands.
 */

/**
 * @brief Callback for ECHO REQUESTs of the ICMP command function.
 *
 * @param parameter The results of the echo request.
 */
static void ICMP_command_send_echo_request_callback(SP_callback_parameter_t parameter) {
	SP_ICMP_echo_request_result_t* echo_request_result = (SP_ICMP_echo_request_result_t*) parameter;

	SP_node_t* node;
	switch(echo_request_result->errno) {
	case SP_ERROR_OUT_OF_MEMORY:
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "ICMP: More echo replies have been received, but they didn't fit into the allocated memory.\n");
		// fall through...
	case SP_ERROR_OK:
		node = (SP_node_t*) list_head(echo_request_result->nodes);
		SP_DEBUG(SP_DEBUG_LEVEL_NOTE, "ICMP: %d node(s) replied to echo request: %s (%d hops)", list_size(echo_request_result->nodes), addr_to_str(node->address), node->hops);
		list_remove(echo_request_result->nodes, node);
		while(!list_empty(echo_request_result->nodes)) {
			node = (SP_node_t*) list_head(echo_request_result->nodes);
			list_remove(echo_request_result->nodes, node);
			if(list_size(echo_request_result->nodes) >= 1) {
				SP_DEBUG(SP_DEBUG_LEVEL_NONE, ", ");
			} else {
				SP_DEBUG(SP_DEBUG_LEVEL_NONE, " and ");
			}
			SP_DEBUG(SP_DEBUG_LEVEL_NONE, "%s (%d hops)", addr_to_str(node->address), node->hops);
		}
		SP_DEBUG(SP_DEBUG_LEVEL_NONE, ".\n");
		break;
	case SP_ERROR_TIMEOUT:
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "ICMP: No replies to echo request.\n");
		break;
	case SP_ERROR_BUSY:
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "ICMP: Echo request in progress.\n");
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "ICMP: Callback with unknown error code 0x%X.\n", echo_request_result->errno);
		break;
	}

	list_free(echo_request_result->nodes);
	SP_FREE(echo_request_result);
}

/**
 * @brief ICMP command function.
 *
 * @param argc Number of command parameters.
 * @param argv Array of strings used as command parameters.
 * @return <code>true</code> if the command was executed successfully,
 * <code>false</code> otherwise.
 */
bool ICMP_command(int argc, const char*const* argv) {
	if(argc == 2) {
		if(strcmp(argv[1], "ICMP_send_nop") == 0) {
			ICMP_send_nop(SP_BROADCAST_ADDR, SP_DEFAULT_TTL);
			return true;
		} else if(strcmp(argv[1], "ICMP_send_echo_request") == 0) {
			SP_ICMP_echo_request_result_t* echo_request_result;
			SP_MALLOC(echo_request_result, SP_ICMP_echo_request_result_t);
			list_alloc(&echo_request_result->nodes);
			ICMP_send_echo_request(SP_BROADCAST_ADDR, SP_DEFAULT_TTL, ICMP_MAX_ECHO_REQUEST_NODES, ICMP_command_send_echo_request_callback, echo_request_result);
			return true;
		} else if(strcmp(argv[1], "ICMP_send_local_echo_request") == 0) {
			SP_ICMP_echo_request_result_t* echo_request_result;
			SP_MALLOC(echo_request_result, SP_ICMP_echo_request_result_t);
			list_alloc(&echo_request_result->nodes);
			ICMP_send_echo_request(SP_BROADCAST_ADDR, 1, ICMP_MAX_ECHO_REQUEST_NODES, ICMP_command_send_echo_request_callback, echo_request_result);
			return true;
		}
	} else if(argc == 3) {
		if(strcmp(argv[1], "ICMP_send_nop") == 0) {
			ICMP_send_nop((SP_addr_t) atoi(argv[2]), SP_DEFAULT_TTL);
			return true;
		} else if(strcmp(argv[1], "ICMP_send_echo_request") == 0) {
			SP_ICMP_echo_request_result_t* echo_request_result;
			SP_MALLOC(echo_request_result, SP_ICMP_echo_request_result_t);
			list_alloc(&echo_request_result->nodes);
			ICMP_send_echo_request((SP_addr_t) atoi(argv[2]), SP_DEFAULT_TTL, ICMP_MAX_ECHO_REQUEST_NODES, ICMP_command_send_echo_request_callback, echo_request_result);
			return true;
		} else if(strcmp(argv[1], "ICMP_send_local_echo_request") == 0) {
			SP_ICMP_echo_request_result_t* echo_request_result;
			SP_MALLOC(echo_request_result, SP_ICMP_echo_request_result_t);
			list_alloc(&echo_request_result->nodes);
			ICMP_send_echo_request((SP_addr_t) atoi(argv[2]), 1, ICMP_MAX_ECHO_REQUEST_NODES, ICMP_command_send_echo_request_callback, echo_request_result);
			return true;
		}
	} else if(argc == 4) {
		if(strcmp(argv[1], "ICMP_send_nop") == 0) {
			ICMP_send_nop((SP_addr_t) atoi(argv[2]), (SP_ttl_t) atoi(argv[3]));
			return true;
		} else if(strcmp(argv[1], "ICMP_send_echo_request") == 0) {
			SP_ICMP_echo_request_result_t* echo_request_result;
			SP_MALLOC(echo_request_result, SP_ICMP_echo_request_result_t);
			list_alloc(&echo_request_result->nodes);
			ICMP_send_echo_request((SP_addr_t) atoi(argv[2]), (SP_ttl_t) atoi(argv[3]), ICMP_MAX_ECHO_REQUEST_NODES, ICMP_command_send_echo_request_callback, echo_request_result);
			return true;
		}
	}
	return false;
}
