/**
 * @file interface.c
 *
 * @brief Service Placement interface implementation.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: interface.c 2567 2008-02-20 15:54:36Z wittenbu $
 */

#include <string.h>

#include "types.h"
#include "vars.h"
#include "interface.h"
#include "interface_internal.h"
#include "util/debug.h"
#include "util/list.h"
#include "util/callback.h"
#include "util/net.h"
#include "icmp-light/icmp-light.h"
#include "rudp-light/rudp-light.h"
#include "slp-light/slp-light.h"
#include "service/service.h"
#include "placement/placement.h"



/**
 * @brief Global initialization function called by the underlying system.
 */
void SP_init() {
	SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Initializing... \n");

	if(SP_LOCAL_ADDR == SP_BROADCAST_ADDR) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Local node is using the broadcast address (%s == %s).\n", addr_to_str(SP_LOCAL_ADDR), addr_to_str(SP_BROADCAST_ADDR));
		abort();
	}

	// Initialize global variables.
	list_alloc(&SP_callbacks);
	SP_callback_time = 0;
	SP_registered_callbacks = 0;
	SP_executed_callbacks = 0;
	SP_sequence_number = 0;
	memset(SP_received_packet_cache, 0, SP_RECEIVED_PACKET_CACHE_SIZE * sizeof(SP_packet_fingerprint_t));
	list_alloc(&SP_network_map);
	SP_direct_neighbors = 0;
	SP_longest_route = 0;
	SP_average_route = 0;
	SP_addr_to_str_buffer_index = 0;

	SP_DEBUG_ERRORs = 0;
	SP_DEBUG_WARNs = 0;

	// Initialize components.
	ICMP_init();
	RUDP_init();
	SLP_init();
	SERVICE_init();
	PLACEMENT_init();
}

/**
 * @brief Global start function called by the underlying system.
 */
void SP_start() {
	SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Starting... \n");

	// Start components.
	ICMP_start();
	RUDP_start();
	SLP_start();
	SERVICE_start();
	PLACEMENT_start();
}

/**
 * @brief Global stop function called by the underlying system.
 */
void SP_stop() {
	SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Stopping...\n");

	// Stop components.
	ICMP_stop();
	RUDP_stop();
	SLP_stop();
	SERVICE_stop();
	PLACEMENT_stop();
}

/**
 * @brief Global finalization function called by the underlying system.
 */
void SP_finalize() {
	SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Destroying...\n");

	calculate_network_statistics();

	if(!list_empty(SP_callbacks)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "There are still %d callback(s) in queue, last one due in %.2f seconds.\n", list_size(SP_callbacks), (((SP_callback_t*) list_tail(SP_callbacks))->time - __SP_time()));
		abort();
	} else if(SP_executed_callbacks != SP_registered_callbacks) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Callback counters don't add up (%d != %d).\n", SP_executed_callbacks, SP_registered_callbacks);
		abort();
	}

	list_free(SP_callbacks);
	list_free(SP_network_map);

	// Destroy components.
	ICMP_finalize();
	RUDP_finalize();
	SLP_finalize();
	SERVICE_finalize();
	PLACEMENT_finalize();
}

/**
 * @brief Internal packet reception function.
 *
 * This function may be called either when packets arrive from the underlying
 * system or if a (possibly encapsulated) packet needs to be processed again.
 *
 * @param header Pointer to the header of the received packet.
 * @param payload Pointer to the payload of the received packet.
 * @param maybe_duplicate Set to true if this packet is a potential duplicate
 * that requires filtering.
 */
void SP_handle_packet(const SP_packet_header_t* header, const SP_packet_payload_t* payload, bool maybe_duplicate) {
	if(header->type == SP_TYPE_NONE) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "Dropping packet with uninitialized type field from node %s...\n", addr_to_str(header->from));
		return;
	} else if(header->ttl == 0) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "Dropping packet with expired TTL from node %s...\n", addr_to_str(header->from));
		return;
	} else if(header->last_hop == SP_LOCAL_ADDR) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "Dropping packet that was last handled by this node (from=%s,seqno=%d)...\n", addr_to_str(header->from), header->seqno);
		return;
	} else if(maybe_duplicate && check_for_duplicate(header)) {
		SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Dropping duplicate packet from node %s (seqno=%d)...\n", addr_to_str(header->from), header->seqno);
		return;
	}

	switch(header->app) {
	case SP_APP_NONE:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "Received packet with uninitialized application field from node %s.\n", addr_to_str(header->from));
		break;
	case SP_APP_ICMP:
		ICMP_recv(header, payload);
		break;
	case SP_APP_RUDP:
		RUDP_recv(header, payload);
		break;
	case SP_APP_SLP:
		SLP_recv(header, payload);
		break;
	case SP_APP_SERVICE:
		SERVICE_recv(header, payload);
		break;
	case SP_APP_PLACEMENT:
		PLACEMENT_recv(header, payload);
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "Received packet for unknown application 0x%X from %s.\n", header->app, addr_to_str(header->from));
	}
}

/**
 * @brief Global packet reception function called by the underlying system.
 *
 * @param header Pointer to the header of the received packet.
 * @param payload Pointer to the payload of the received packet.
 */
void SP_recv(const SP_packet_header_t* header, const SP_packet_payload_t* payload) {
	add_node_to_network_map(header->from, header->hops, header->last_hop);
	SP_handle_packet(header, payload, (header->to == SP_BROADCAST_ADDR));
}

/**
 * @brief Global callback trigger function called by the underlying system.
 */
void SP_execute_callback() {
	callback_expire();
}

/**
 * @brief Global command function called by the underlying system.
 *
 * @param argc Length of the argument array.
 * @param argv Array of argument strings.
 * @return <code>true</code> if the command was processed successfully,
 * <code>false</code> otherwise.
 */
bool SP_command(int argc, const char*const* argv) {
	if(argc >= 2) {
		SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Received command \"%s\"...\n", argv[1]);
	} else {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "Received command without command string...\n");
	}

	// Call into components.
	if(ICMP_command(argc, argv))
		return true;
	if(RUDP_command(argc, argv))
		return true;
	if(SLP_command(argc, argv))
		return true;
	if(SERVICE_command(argc, argv))
		return true;
	if(PLACEMENT_command(argc, argv))
		return true;
	return false;
}

/**
 * @brief Calculate the maximal payload size.
 *
 * @return Maximal payload size in bytes.
 */
size_t SP_max_payload_size() {
	return (SP_MTU - sizeof(SP_packet_header_t));
}



/**
 * @brief Wrapper function around low-level output function.
 *
 * @param format Format string.
 * @param ... Variable number of arguments matching the format string.
 */
void SP_printf(const char *format, ...) {
	va_list argp;
	va_start(argp, format);
	__SP_vprintf(format, argp);
	va_end(argp);
}

/**
 * @brief Wrapper function around low-level output function.
 *
 * @param format Format string.
 * @param ap Variable length argument structure.
 */
void SP_vprintf(const char *format, va_list ap) {
	__SP_vprintf(format, ap);
}

/**
 * @brief Wrapper function around low-level random function.
 *
 * @param max Maximal value.
 * @return Random value between 0 and max.
 */
double SP_random(double max) {
	return __SP_random(max);
}

/**
 * @brief Wrapper function around low-level time function.
 *
 * @return Current time.
 */
SP_time_t SP_time() {
	return __SP_time();
}

/**
 * @brief Wrapper function around low-level send function.
 *
 * @param header Header of packet to send.
 * @param payload Payload of packet to send.
 */
void SP_send(const SP_packet_header_t* header, const SP_packet_payload_t* payload) {
	if(header == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Trying to send packet without header.\n");
		abort();
	} else if((header->payload_length > 0) && (payload == NULL)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Trying to send non-empty packet without payload.\n");
		abort();
	} else if((sizeof(SP_packet_header_t) + header->payload_length) > SP_MTU) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Trying to send packet exceeding the MTU (%d > %d).\n", (sizeof(SP_packet_header_t) + header->payload_length), SP_MTU);
		abort();
	}

	if(header->to != SP_LOCAL_ADDR)
		__SP_send(header, payload);
	else
		SP_handle_packet(header, payload, false);
}

/**
 * @brief Wrapper function around low-level callback activation function.
 *
 * @param time Time at which the next callback needs to be activated.
 */
void SP_activate_callback(SP_time_t time) {
	if(time < SP_time()) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "Trying to activate callback in the past (%f < %f).\n", time, SP_time());
		return;
	}
	SP_callback_time = time;
	__SP_activate_callback(time);
}
