/**
 * @file interface.h
 *
 * @brief Service Placement interface (header).
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: interface.h 2548 2008-02-13 19:20:58Z wittenbu $
 */

#ifndef INTERFACE_H
#define INTERFACE_H

#include <stdarg.h>

#include "types.h"

// Interface functions to be called by underlying system.
void SP_init();
void SP_start();
void SP_stop();
void SP_finalize();
void SP_recv(const SP_packet_header_t* header, const SP_packet_payload_t* payload);
void SP_execute_callback();
bool SP_command(int argc, const char*const* argv);

// Wrappers around functions provided by the underlying system.
void SP_vprintf(const char *format, va_list ap);
double SP_random(double max);
SP_time_t SP_time();
void SP_send(const SP_packet_header_t* header, const SP_packet_payload_t* payload);
void SP_activate_callback(SP_time_t time);

// Functions provided by the underlying system.

/**
 * @brief Platform specific low-level output function.
 *
 * @param format Format string.
 * @param ap Variable length argument structure.
 */
void __SP_vprintf(const char *format, va_list ap);

/**
 * @brief Platform specific low-level random function.
 *
 * @param max Maximal value.
 * @return Random value between 0 and max.
 */
double __SP_random(double max);

/**
 * @brief Platform specific low-level time function.
 *
 * @return Current time.
 */
SP_time_t __SP_time();

/**
 * @brief Platform specific low-level send function.
 *
 * @param header Header of packet to send.
 * @param payload Payload of packet to send.
 */
void __SP_send(const SP_packet_header_t* header, const SP_packet_payload_t* payload);

/**
 * @brief Platform specific low-level callback activation function.
 *
 * @param time Time at which the next callback needs to be activated.
 */
void __SP_activate_callback(SP_time_t time);

#endif	// INTERFACE_H
