/**
 * @file main.c
 *
 * @brief Service Placement main file.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: main.c 2588 2008-03-06 11:44:37Z wittenbu $
 */

/**
 * @mainpage
 *
 * The problem of service placement in ad hoc networks can be stated as follows:
 * Given a physical network topology and service demands of client nodes, adapt
 * the number and location of services in the network over time such that the
 * service demands are met at a minimal cost. The cost function may include
 * metrics such as network traffic, energy expenditure or service-dependant
 * Quality of Service (QoS) metrics (availability, latency, etc.).
 *
 * The purpose of this framework is to evaluate different approaches to service
 * placement in ad hoc networks. It is designed to run on different platforms
 * ranging from network simulators to major operating systems and embedded
 * devices.
 *
 * For more information, please have a look at the project webpage at
 * http://cst.mi.fu-berlin.de/projects/ServicePlacement/.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 */

#include <error.h>
#include <netdb.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <netinet/in.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>

#include "types.h"
#include "vars.h"
#include "interface.h"
#include "util/debug.h"
#include "util/callback.h"
#include "util/net.h"
#include "service/service.h"

#include <errno.h>	// Include after service.h due to conflicting macros and typedefs.



/**
 * @brief Status of the application.
 */
bool running;

/**
 * @brief Size of buffer for received packets in bytes.
 */
#define RECV_BUFFER_SIZE	1500

/**
 * @brief Buffer for received packets in bytes.
 */
unsigned char recv_buffer[RECV_BUFFER_SIZE];

/**
 * @brief Time interval at which to insert a mark into logfile in seconds.
 */
#define LOG_MARK_INTERVAL	1



/**
 * @brief Initialization of the application.
 *
 * Seed the random number generator, setup and configure a socket, establish
 * the local address and initial
 */
static void init() {
	// Seed random number generator.
	unsigned int seed = (unsigned int) (SP_time() * 1000000);
	SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Setting random seed to %d...\n", seed);
	srandom(seed);

	// Setup socket.
	struct protoent *udp = getprotobyname("UDP");
	if(udp == NULL) {
		int errsv = errno;
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Error while retrieving protocol information for UDP... (%d)\n", errsv);
		error(1, errsv, NULL);
	}
	if((SP_udp_socket = socket(PF_INET, SOCK_DGRAM, udp->p_proto)) < 0) {
		int errsv = errno;
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Error while creating UDP socket... (%d)\n", errsv);
		error(1, errsv, NULL);
	}
	struct sockaddr_in addr;
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(SP_DEFAULT_UDP_PORT);
	if(bind(SP_udp_socket, (struct sockaddr*) &addr, sizeof(struct sockaddr_in)) < 0) {
		int errsv = errno;
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Error while binding socket... (%d)\n", errsv);
		error(1, errsv, NULL);
	}
	// Enable support for broadcasting.
	int opt = 1;
	if(setsockopt(SP_udp_socket, SOL_SOCKET, SO_BROADCAST, &opt, sizeof(int)) < 0) {
		int errsv = errno;
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Error while setting socket options... (%d)\n", errsv);
		error(1, errsv, NULL);
	}

	// Find out local IP address
	// NOTE: This is Linux-specific code.
	SP_local_addr = 0;
	int i = 1;
	while(true) {
		// Adapted from http://www.hungry.com/~alves/local-ip-in-C.html.
		struct ifreq ifr;
		struct sockaddr_in *sin = (struct sockaddr_in*) &ifr.ifr_addr;
		ifr.ifr_ifindex = i++;
		if(ioctl(SP_udp_socket, SIOCGIFNAME, &ifr) < 0)
			break;
		if(
			(ioctl(SP_udp_socket, SIOCGIFADDR, &ifr) < 0)
			|| (strncmp(inet_ntoa(sin->sin_addr), "127.0.0.1", 10) == 0)
		)
			continue;
		SP_local_addr = sin->sin_addr.s_addr;
	}
	if(SP_local_addr == 0) {
		int errsv = errno;
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Cannot establish local IP address. (%d)\n", errsv);
		error(1, errsv, NULL);
	}
}

/**
 * @brief Handle incoming packets.
 *
 * Retrieve meta-information and hand packet up to components.
 *
 */
static void handle_packet() {
	struct sockaddr_in addr;
	socklen_t addrlen = sizeof(struct sockaddr_in);
	ssize_t size = recvfrom(SP_udp_socket, recv_buffer, RECV_BUFFER_SIZE, 0, (struct sockaddr*) &addr, &addrlen);

	if(size > 0) {
		if(size < ((int) sizeof(SP_packet_header_t))) {
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Received %d byte(s) but expected at least %d bytes.\n", size, sizeof(SP_packet_header_t));
			abort();
		}
		SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Received a packet of size %d from %s:%d.\n", size, addr_to_str(addr.sin_addr.s_addr), ntohs(addr.sin_port));

		SP_packet_header_t* header = (SP_packet_header_t*) recv_buffer;
		header->hops = 0;				// TODO: This actually depends on the routing protocol.
		header->last_hop = addr.sin_addr.s_addr;	// TODO: This actually depends on the routing protocol.
		SP_packet_payload_t* payload = NULL;
		if(header->payload_length > 0)
			payload = recv_buffer + sizeof(SP_packet_header_t);
		SP_recv(header, payload);

		addrlen = sizeof(struct sockaddr_in);	// Reset size of addr variable.
	} else if(size == 0) {	// Can't happen.
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Non-existant connection has been shutdown by peer (recvfrom() returned 0).\n");
		abort();
	} else {		// size is -1.
		int errsv = errno;
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Error while receiving packet. (%d)\n", errsv);
		error(1, errsv, NULL);
	}
}

/**
 * @brief Main loop iteration.
 *
 * Wait for either the next callback or for a packet to arrive.
 */
static void loop_iteration() {
	fd_set incoming_sockets;
	struct timeval timeout;
	SP_time_t time_until_callback;
	bool waiting_for_callback;

	// Check whether callbacks are scheduled.
	waiting_for_callback = false;
	timeout.tv_sec = LOG_MARK_INTERVAL;
	timeout.tv_usec = 0;
	if(SP_callback_time > 0) {
		time_until_callback = SP_callback_time - SP_time();
		if(time_until_callback < LOG_MARK_INTERVAL) {
			waiting_for_callback = true;
			timeout.tv_sec = time_until_callback;
			timeout.tv_usec = (time_until_callback - ((SP_time_t) timeout.tv_sec)) * 1000000;
		}
	}

	// Wait for either a packet to arrive or for the timeout to expire.
	FD_ZERO(&incoming_sockets);
	FD_SET(SP_udp_socket, &incoming_sockets);
	int ready = select(SP_udp_socket + 1, &incoming_sockets, NULL, NULL, &timeout);
	if((ready == 1) && FD_ISSET(SP_udp_socket ,&incoming_sockets)) {
		handle_packet();
	} else if(ready == 0) {
		if(waiting_for_callback)
			callback_expire(NULL);
		else if(SP_callback_time > 0)
			SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "-- MARK -- (next callback in %.2f seconds)\n", SP_callback_time - SP_time());
		else
			SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "-- MARK --\n");
	} else {	// ready is -1.
		int errsv = errno;
		if(errsv != EINTR) {
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Error waiting for packet. (%d)\n", errsv);
			error(1, errsv, NULL);
		}
	}
}

/**
 * @brief Prevents future iterations of the main loop.
 *
 * This function is called in response to a SIGINT signal.
 */
static void stop() {
	running = false;
}

/**
 * @brief The main function.
 *
 * Initialize and start all components, run the main loop and on shutdown stop
 * and finalize all components.
 *
 * @param argc Number of command line arguments.
 * @param argv Array of command line arguments
 */
int main(int argc, char **argv) {
	init();
	signal(SIGINT, stop);

	// Initialize and start components.
	SP_init();
	SP_start();

// FIXME: Do something.
SERVICE_create_service("srv_A", 1024);
SERVICE_request_data("srv_A", 0x0000, 1024);
//SERVICE_request_stream("srv_A", 1024, 10, 30);

	// Main loop.
	running = true;
	while(running)
		loop_iteration();

	// Stop and finalize components.
	SP_stop();
	SP_finalize();

	// Close socket.
	close(SP_udp_socket);

	return 0;
}
