/**
 * @file service_server.c
 *
 * @brief Service - Server.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service_server.c 2548 2008-02-13 19:20:58Z wittenbu $
 */

#include <string.h>

#include "../types.h"
#include "../vars.h"
#include "../interface.h"
#include "../interface_internal.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "../rudp-light/rudp-light.h"
#include "../placement/placement.h"
#include "service.h"
#include "service_types.h"
#include "service_vars.h"
#include "service_server.h"
#include "service_management.h"

/**
 * @brief Maximal number of served streams.
 */
#define SERVICE_MAX_SERVED_STREAMS	64

/*
 * Packet handling functions.
 */

/**
 * @brief Send a SERVICE DATA ITEM packet.
 *
 * Also send redirection information if the service is redirecting.
 *
 * @param to Address of destination node.
 * @param service Pointer to requested service.
 * @param offset Offset in service data.
 * @param size Requested service data size in bytes.
 * @param start_time Start time of this data request in seconds.
 */
static void send_SERVICE_SERVER_DATA_ITEM(SP_addr_t to, const SP_SERVICE_service_t* service, u_int16_t offset, size_t size, SP_time_t start_time) {
	SP_packet_header_t header;
	SP_SERVICE_SERVER_DATA_ITEM_packet_t data_item;

	if(size > service->data_size) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Requested too much data (%d > %d), request size reduced to %d.\n", size, service->data_size, service->data_size);
		size = service->data_size;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Sending %d byte(s) SERVER DATA ITEM for service \"%s\" (0x%X...) to node %s...\n", size, service->id, (u_int8_t) *service->data, addr_to_str(to));
	init_header(&header);
	header.to = to;
	header.app = SP_APP_SERVICE;
	header.type = SP_TYPE_SERVICE_SERVER_DATA_ITEM;
	header.payload_length = sizeof(SP_SERVICE_SERVER_DATA_ITEM_packet_t) + size;

	strncpy(data_item.service_id, service->id, SP_SERVICE_ID_LENGTH);
	data_item.start_time = start_time;
	if(service->state == SP_SERVICE_STATE_REDIRECTING) {
		if(service->redirect_addr == SP_BROADCAST_ADDR) {
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Service is redirecting, but address is set to broadcast.\n");
			abort();
		}
		data_item.redirect_addr = service->redirect_addr;
	} else {
		data_item.redirect_addr = SP_BROADCAST_ADDR;
	}

	SP_packet_payload_t* payload;
	SP_MALLOC_JOIN(payload, SP_packet_payload_t, &data_item, sizeof(SP_SERVICE_SERVER_DATA_ITEM_packet_t), (service->data + offset), size);
	SP_send(&header, payload);
	SP_FREE(payload);
}

/**
 * @brief Send a SERVICE STREAM ITEM packet.
 *
 * Also send redirection information if the service is redirecting.
 *
 * @param to Address of destination node.
 * @param service Pointer to requested service.
 * @param offset Offset in service data.
 * @param size Requested service data size in bytes.
 */
static void send_SERVICE_SERVER_STREAM_ITEM(SP_addr_t to, const SP_SERVICE_service_t* service, u_int16_t offset, size_t size) {
	SP_packet_header_t header;
	SP_SERVICE_SERVER_STREAM_ITEM_packet_t stream_item;

	if(size > service->data_size) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Requested too much data (%d > %d), request size reduced to %d.\n", size, service->data_size, service->data_size);
		size = service->data_size;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Sending %d byte(s) SERVER STREAM ITEM for service \"%s\" (0x%X...) to node %s...\n", size, service->id, (u_int8_t) *service->data, addr_to_str(to));
	init_header(&header);
	header.to = to;
	header.app = SP_APP_SERVICE;
	header.type = SP_TYPE_SERVICE_SERVER_STREAM_ITEM;
	header.payload_length = sizeof(SP_SERVICE_SERVER_STREAM_ITEM_packet_t) + size;

	strncpy(stream_item.service_id, service->id, SP_SERVICE_ID_LENGTH);
	if(service->state == SP_SERVICE_STATE_REDIRECTING) {
		if(service->redirect_addr == SP_BROADCAST_ADDR) {
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Service is redirecting, but address is set to broadcast.\n");
			abort();
		}
		stream_item.redirect_addr = service->redirect_addr;
	} else {
		stream_item.redirect_addr = SP_BROADCAST_ADDR;
	}

	SP_packet_payload_t* payload;
	SP_MALLOC_JOIN(payload, SP_packet_payload_t, &stream_item, sizeof(SP_SERVICE_SERVER_STREAM_ITEM_packet_t), (service->data + offset), size);
	SP_send(&header, payload);
	SP_FREE(payload);
}

/**
 * @brief Send SERVICE NOT AVAILABLE packet.
 *
 * @param to Address of destination node.
 * @param service_id ID of the requested service.
 * @param redirect_addr Address to redirect future service requests to.
 */
static void send_SERVICE_SERVER_NOT_AVAILABLE(SP_addr_t to, const char* service_id, SP_addr_t redirect_addr) {
	SP_packet_header_t header;
	SP_SERVICE_SERVER_NOT_AVAILABLE_packet_t not_available;

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Sending SERVER NOT AVAILABLE for service \"%s\" to node %s...\n", service_id, addr_to_str(to));
	init_header(&header);
	header.to = to;
	header.app = SP_APP_SERVICE;
	header.type = SP_TYPE_SERVICE_SERVER_NOT_AVAILABLE;
	header.payload_length = sizeof(SP_SERVICE_SERVER_NOT_AVAILABLE_packet_t);
	strncpy(not_available.service_id, service_id, SP_SERVICE_ID_LENGTH);
	not_available.redirect_addr = redirect_addr;
	RUDP_send(&header, (SP_packet_payload_t*) &not_available, NULL, NULL);
}

/**
 * @brief Handle incoming SERVICE DATA REQUEST packet.
 *
 * Send a SERVICE DATA ITEM packet if the service is available or a SERVICE NOT
 * AVAILABLE packet otherwise.
 *
 * @param header Header of incoming packet.
 * @param request Data request information.
 */
void handle_SERVICE_CLIENT_DATA_REQUEST(const SP_packet_header_t* header, const SP_SERVICE_CLIENT_DATA_REQUEST_packet_t* request) {
	SP_SERVICE_service_t* service = get_service(request->service_id);
	if(service == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Received CLIENT DATA REQUEST from node %s, but service \"%s\" is not available.\n", addr_to_str(header->from), request->service_id);
		send_SERVICE_SERVER_NOT_AVAILABLE(header->from, request->service_id, SP_BROADCAST_ADDR);
		return;
	} else if(
		(service->state != SP_SERVICE_STATE_RUNNING)
		&& (service->state != SP_SERVICE_STATE_REPLICATING)
		&& (service->state != SP_SERVICE_STATE_REDIRECTING)
	) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Received CLIENT DATA REQUEST from node %s, but service \"%s\" is not active.\n", addr_to_str(header->from), request->service_id);
		if(service->state == SP_SERVICE_STATE_STOPPED)
			send_SERVICE_SERVER_NOT_AVAILABLE(header->from, request->service_id, service->redirect_addr);
		return;
	} else if(service->data_size < (request->offset + request->size)) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Received CLIENT DATA REQUEST for %d byte(s) at 0x%X of service \"%s\" from node %s, but request size exceeds service data (%d < (%d + %d)).\n", request->size, request->offset, request->service_id, addr_to_str(header->from), service->data_size, request->offset, request->size);
		return;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Received CLIENT DATA REQUEST for %d byte(s) at 0x%X of service \"%s\" from node %s.\n", request->size, request->offset, request->service_id, addr_to_str(header->from));
	send_SERVICE_SERVER_DATA_ITEM(header->from, service, request->offset, request->size, request->start_time);
	PLACEMENT_stat_service_request(service, header->from, header->last_hop, header->hops, request->size);
}

/**
 * @brief Handle incoming SERVICE DATA UPDATE packet.
 *
 * Update the service data if the srevice is available or send a SERVICE NOT
 * AVAILABLE packet otherwise.
 *
 * @param header Header of incoming packet.
 * @param payload Payload of incoming packet containing the update information
 * and the new service data.
 */
void handle_SERVICE_CLIENT_DATA_UPDATE(const SP_packet_header_t* header, const SP_packet_payload_t* payload) {
	SP_SERVICE_CLIENT_DATA_UPDATE_packet_t update;
	char* update_data;
	SP_SERVICE_service_t* service;

	if(header->payload_length < sizeof(SP_SERVICE_CLIENT_DATA_UPDATE_packet_t)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: Received CLIENT DATA UPDATE packet with wrong payload length (%d < %d).\n", header->payload_length, sizeof(SP_SERVICE_CLIENT_DATA_UPDATE_packet_t));
		abort();
	}

	size_t update_size = header->payload_length - sizeof(SP_SERVICE_CLIENT_DATA_UPDATE_packet_t);
	SP_MALLOC_ARRAY(update_data, char, update_size);
	SP_SPLIT(&update, sizeof(SP_SERVICE_CLIENT_DATA_UPDATE_packet_t), update_data, update_size, payload);

	service = get_service(update.service_id);
	if(service == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Received CLIENT DATA UPDATE from node %s, but service \"%s\" is not available.\n", addr_to_str(header->from), update.service_id);
		send_SERVICE_SERVER_NOT_AVAILABLE(header->from, update.service_id, SP_BROADCAST_ADDR);
		return;
	} else if(service->state != SP_SERVICE_STATE_RUNNING) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Received CLIENT DATA UPDATE from node %s, but service \"%s\" is not active.\n", addr_to_str(header->from), update.service_id);
		if(
			(service->state == SP_SERVICE_STATE_REDIRECTING)
			|| (service->state == SP_SERVICE_STATE_STOPPED)
		) {
			send_SERVICE_SERVER_NOT_AVAILABLE(header->from, update.service_id, service->redirect_addr);
		}
		return;

	// NOTE: We silently ignore DATA UPDATES that arrive while a service is
	// replicating. This may cause data inconsistencies.

	} else if(service->data_size < (update.offset + update_size)) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Received CLIENT DATA UPDATE for %d byte(s) at 0x%X of service \"%s\" from node %s, but update size exceeds service data (%d < (%d + %d)).\n", update_size, update.offset, update.service_id, addr_to_str(header->from), service->data_size, update.offset, update_size);
		return;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Received CLIENT DATA UPDATE for %d byte(s) at 0x%X of service \"%s\" (0x%X...) from node %s.\n", update_size, update.offset, update.service_id, (u_int8_t) *update_data, addr_to_str(header->from));
	update_service_data(service, update.offset, update_size, update_data);
	PLACEMENT_stat_service_request(service, header->from, header->last_hop, header->hops, update_size);

	// NOTE: A this point we would have to take care of synchronizing with
	// other service instances.
}

/**
 * @brief Process an active stream.
 *
 * Send a SERVICE STREAM ITEM if the service is still active, possibly
 * reschedule until the next packet is due.
 *
 * @param parameter Pointer to stream to be processed.
 */
static void process_SERVICE_CLIENT_STREAM_REQUEST(SP_callback_parameter_t parameter) {
	SP_SERVICE_stream_t* stream = (SP_SERVICE_stream_t*) parameter;
	SP_SERVICE_service_t* service = get_service(stream->service_id);

	if(service == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Trying to stream service \"%s\" to node %s, but service is not available.\n", stream->service_id, addr_to_str(stream->to));
		send_SERVICE_SERVER_NOT_AVAILABLE(stream->to, stream->service_id, SP_BROADCAST_ADDR);
		list_remove(SP_SERVICE_served_streams, stream);
		SP_FREE(stream);
		return;
	} else if(
		(service->state != SP_SERVICE_STATE_RUNNING)
		&& (service->state != SP_SERVICE_STATE_REPLICATING)
		&& (service->state != SP_SERVICE_STATE_REDIRECTING)
	) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Trying to stream service \"%s\" to node %s, but service is not active.\n", stream->service_id, addr_to_str(stream->to));
		if(service->state == SP_SERVICE_STATE_STOPPED)
			send_SERVICE_SERVER_NOT_AVAILABLE(stream->to, stream->service_id, service->redirect_addr);
		list_remove(SP_SERVICE_served_streams, stream);
		SP_FREE(stream);
		return;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Processing stream of service \"%s\" to node %s...\n", stream->service_id, addr_to_str(stream->to));
	send_SERVICE_SERVER_STREAM_ITEM(stream->to, service, stream->offset, stream->size);
	if((SP_time() + stream->interval) <= stream->end_time) {
		callback_add(process_SERVICE_CLIENT_STREAM_REQUEST, stream, stream->interval);
	} else {
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SERVICE: Stream of service \"%s\" to node %s has ended.\n", stream->service_id, addr_to_str(stream->to));
		list_remove(SP_SERVICE_served_streams, stream);
		SP_FREE(stream);
	}
}

/**
 * @brief Check whether a two streams match.
 *
 * @param item Stream to check.
 * @param cmp_item Reference stream.
 * @return <code>true</code> if the two streams match, <code>false</code>
 * otherwise.
 */
static bool matches_stream(const void* item, const void* cmp_item) {
	SP_SERVICE_stream_t* stream = (SP_SERVICE_stream_t*) item;
	SP_SERVICE_stream_t* cmp_stream = (SP_SERVICE_stream_t*) cmp_item;
	return (
		(stream->to == cmp_stream->to)
		&& (strcmp(stream->service_id, cmp_stream->service_id) == 0)
		&& (stream->offset == cmp_stream->offset)
		&& (stream->size == cmp_stream->size)
		&& (stream->interval == cmp_stream->interval)
	);
}

/**
 * @brief Handle incoming SERVICE STREAM REQUEST packet.
 *
 * Schedule a stream for processing if the service is available and additional
 * streams can be supported, send a SERVICE NOT AVAILABLE packet otherwise.
 *
 * @param header Header of incoming packet.
 * @param request Stream request information.
 */
void handle_SERVICE_CLIENT_STREAM_REQUEST(const SP_packet_header_t* header, const SP_SERVICE_CLIENT_STREAM_REQUEST_packet_t* request) {
	SP_SERVICE_service_t* service = get_service(request->service_id);
	if(service == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Received CLIENT STREAM REQUEST from node %s, but service \"%s\"is not available.\n", addr_to_str(header->from), request->service_id);
		send_SERVICE_SERVER_NOT_AVAILABLE(header->from, request->service_id, SP_BROADCAST_ADDR);
		return;
	} else if((service->state != SP_SERVICE_STATE_RUNNING) && (service->state != SP_SERVICE_STATE_REPLICATING)) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Received CLIENT STREAM REQUEST from node %s, but service \"%s\"is not active.\n", addr_to_str(header->from), request->service_id);
		if(service->state == SP_SERVICE_STATE_STOPPED)
			send_SERVICE_SERVER_NOT_AVAILABLE(header->from, request->service_id, service->redirect_addr);
		return;
	} else if(list_size(SP_SERVICE_served_streams) >= SERVICE_MAX_SERVED_STREAMS) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: No empty slots for stream processing.\n");
		send_SERVICE_SERVER_NOT_AVAILABLE(header->from, request->service_id, SP_BROADCAST_ADDR);
		return;
	} else if(service->data_size < (request->offset + request->size)) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Received CLIENT STREAM REQUEST for %d byte(s) at 0x%X of service \"%s\" from node %s, but request size exceeds service data (%d < (%d + %d)).\n", request->size, request->offset, request->service_id, addr_to_str(header->from), service->data_size, request->offset, request->size);
		return;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Received CLIENT STREAM REQUEST for %d byte(s) at 0x%X of service \"%s\" at interval %.2fs for a duration of %.3fs from node %s.\n", request->size, request->offset, request->service_id, request->interval, request->duration, addr_to_str(header->from));
	SP_SERVICE_stream_t cmp_stream;
	cmp_stream.to = header->from;
	strncpy(cmp_stream.service_id, request->service_id, SP_SERVICE_ID_LENGTH);
	cmp_stream.offset = request->offset;
	cmp_stream.size = request->size;
	cmp_stream.interval = request->interval;
	if(list_find(SP_SERVICE_served_streams, &cmp_stream, matches_stream)) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: A matching CLIENT STREAM REQUEST is already being served, ignoring request.\n");
		return;
	}

	// Initialize stream.
	SP_SERVICE_stream_t* stream;
	SP_MALLOC(stream, SP_SERVICE_stream_t);
	stream->to = header->from;
	strncpy(stream->service_id, request->service_id, SP_SERVICE_ID_LENGTH);
	stream->offset = request->offset;
	stream->size = request->size;
	stream->interval = request->interval;
	stream->end_time = request->duration + SP_time();
	list_add(SP_SERVICE_served_streams, stream);
	callback_add(process_SERVICE_CLIENT_STREAM_REQUEST, stream, 0);

	unsigned int expected_stream_size = request->size * ((unsigned int) (request->duration / request->interval)) + 1;
	PLACEMENT_stat_service_request(service, header->from, header->last_hop, header->hops, expected_stream_size);
}
