/**
 * @file service_util.c
 *
 * @brief Service utility functions.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: service_util.c 2585 2008-03-05 13:28:49Z wittenbu $
 */

#include <string.h>

#include "../types.h"
#include "../vars.h"
#include "../interface.h"
#include "../interface_internal.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "../slp-light/slp-light_types.h"
#include "service.h"
#include "service_types.h"
#include "service_vars.h"
#include "service_util.h"


/**
 * @brief Check whether two service IDs match.
 *
 * @param service_id_1 First service ID.
 * @param service_id_2 Seconds service ID.
 * @return <code>true</code> if the two service IDs are equal,
 * <code>false</code> otherwise.
 */
bool same_service_id(const char* service_id_1, const char* service_id_2) {
	if(strncmp(service_id_1, service_id_2, SP_SERVICE_ID_LENGTH) == 0)
		return true;
	else
		return false;
}

/**
 * @brief Relative weight of the hop metric when selecting the best server.
 */
#define SERVICE_HOP_MULTIPLIER		0.5

/**
 * @brief Relative weight of the delay metric when selecting the best server.
 */
#define SERVICE_DELAY_MULTIPLIER	0.5

/**
 * @brief Select the best server from several found service instances.
 *
 * @param service_instances List of service instances.
 * @return Address of the best server.
 */
SP_addr_t selectBestServer(const SP_list_t* service_instances) {
	if(list_empty(service_instances)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SERVICE: No service instances available for selection.\n");
		abort();
	}

	SP_addr_t server = SP_BROADCAST_ADDR;
	float best_service_metric = -1;
	unsigned int i;
	for(i = 0; i < list_size(service_instances); i++) {
		SP_SLP_service_instance_t* service_instance = (SP_SLP_service_instance_t*) list_get(service_instances, i);
		float metric = 0;
		if((service_instance->hops > 0) && (service_instance->delay > 0)) {
			metric += SERVICE_HOP_MULTIPLIER * (((float) SP_EXPECTED_NETWORK_DIAMETER) / ((float) service_instance->hops));
			metric += SERVICE_DELAY_MULTIPLIER * (((float) SP_EXPECTED_MULTI_HOP_RTT) / service_instance->delay);
		}
		if(metric > best_service_metric) {
			server = service_instance->server;
			best_service_metric = metric;
		}
	}
	return server;
}



/*
 * Client-side cache for results of SLP queries.
 */

/**
 * @brief Size of cache services cache.
 */
#define SERVICE_SERVER_CACHE_SIZE	SP_MAX_SERVICES

/**
 * @brief Maximal age of a cache entry.
 *
 * Note that the age of an entry is not a concept of time, but just a counter
 * that may be incremented given the appropriate circumstances.
 */
#define SERVICE_MAX_CACHE_ENTRY_AGE	16

/**
 * @brief Check whether two cache entries have the same service ID.
 *
 * @param item Cache entry to check.
 * @param cmp_item Reference cache entry.
 * @return <code>true</code> if the two cache entries have the same service ID,
 * <code>false</code> otherwise.
 */
static bool matches_entry_service_id(const void* item, const void* cmp_item) {
	SP_SERVICE_service_cache_entry_t* entry = (SP_SERVICE_service_cache_entry_t*) item;
	SP_SERVICE_service_cache_entry_t* cmp_entry = (SP_SERVICE_service_cache_entry_t*) cmp_item;
	return same_service_id(entry->service_id, cmp_entry->service_id);
}

/**
 * @brief Set the entry in the service cache to a server address.
 *
 * @param service_id Service ID of the cached service.
 * @param server Address of the server for this service.
 */
void cache_set(const char* service_id, SP_addr_t server) {
	if(list_size(SP_SERVICE_service_cache) >= SERVICE_SERVER_CACHE_SIZE) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Server cache is full.\n");
		return;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Setting cache entry for service \"%s\" to node %s...\n", service_id, addr_to_str(server));
	SP_SERVICE_service_cache_entry_t cmp_entry;
	strncpy(cmp_entry.service_id, service_id, SP_SERVICE_ID_LENGTH);
	SP_SERVICE_service_cache_entry_t* entry = (SP_SERVICE_service_cache_entry_t*) list_find(SP_SERVICE_service_cache, &cmp_entry, matches_entry_service_id);
	if(entry == NULL) {
		SP_MALLOC(entry, SP_SERVICE_service_cache_entry_t);
		strncpy(entry->service_id, service_id, SP_SERVICE_ID_LENGTH);
		list_add(SP_SERVICE_service_cache, entry);
	}
	entry->server = server;
	entry->age = 0;
	entry->reachable = true;
}

/**
 * @brief Remove a service from the cache.
 *
 * @param service_id ID of the service that is to be removed.
 */
void cache_remove(const char* service_id) {
	SP_SERVICE_service_cache_entry_t cmp_entry;
	strncpy(cmp_entry.service_id, service_id, SP_SERVICE_ID_LENGTH);
	SP_SERVICE_service_cache_entry_t* entry = (SP_SERVICE_service_cache_entry_t*) list_find(SP_SERVICE_service_cache, &cmp_entry, matches_entry_service_id);

	if((entry != NULL) && (!entry->reachable)) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Trying to remove cache entry for service \"%s\" that is not reachable...\n", service_id);
		return;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Removing cache entry for service \"%s\"...\n", service_id);
	// We need to check whether the entry was already removed from the
	// list, e.g. by a DATA REQUEST overlapping with a SERVER NOT AVAILABLE.
	if(entry != NULL) {
		list_remove(SP_SERVICE_service_cache, entry);
		SP_FREE(entry);
	}
}

/**
 * @brief Lookup a service in the cache.
 *
 * @param service_id ID of the service to lookup.
 * @param server Pointer to the server address if SP_ERROR_OK is returned,
 * Pointer to SP_BROADCAST_ADDR otherwise.
 * @return SP_ERROR_NOT_FOUND if the service in not in the cache,
 * SP_ERROR_TIMEOUT if the service is currently not reachable, or SP_ERROR_OK
 * if a valid cache entry has been found.
 */
u_int8_t cache_lookup(const char* service_id, SP_addr_t* server) {
	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Looking up cache entry for service \"%s\"...\n", service_id);
	SP_SERVICE_service_cache_entry_t cmp_entry;
	strncpy(cmp_entry.service_id, service_id, SP_SERVICE_ID_LENGTH);
	SP_SERVICE_service_cache_entry_t* entry = (SP_SERVICE_service_cache_entry_t*) list_find(SP_SERVICE_service_cache, &cmp_entry, matches_entry_service_id);
	if(entry == NULL) {
		*server = SP_BROADCAST_ADDR;
		return SP_ERROR_NOT_FOUND;
	} else if(entry->reachable) {
		*server = entry->server;
		return SP_ERROR_OK;
	} else {
		*server = SP_BROADCAST_ADDR;
		return SP_ERROR_TIMEOUT;
	}
}

/**
 * @brief Increment the age of a cache entry.
 *
 * If this causes the age of the entry to exceed a limit, the entry is removed
 * from the cache.
 *
 * @param service_id ID of the service whose cache entry is to be incremented.
 */
void cache_increment_age(const char* service_id) {
	SP_SERVICE_service_cache_entry_t cmp_entry;
	strncpy(cmp_entry.service_id, service_id, SP_SERVICE_ID_LENGTH);
	SP_SERVICE_service_cache_entry_t* entry = (SP_SERVICE_service_cache_entry_t*) list_find(SP_SERVICE_service_cache, &cmp_entry, matches_entry_service_id);
	if(entry == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Trying to increment age of cache entry for service \"%s\", but entry doesn't exists...\n", service_id);
		return;
	} else if(!entry->reachable) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Trying to increment age of cache entry for service \"%s\", but service is not reachable...\n", service_id);
		return;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Incrementing age of cache entry for service \"%s\" pointing to node %s...\n", service_id, addr_to_str(entry->server));
	entry->age++;
	if(entry->age > SERVICE_MAX_CACHE_ENTRY_AGE) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Cache entry for service \"%s\" pointing to node %s exceeds age limit, removing...\n", service_id, addr_to_str(entry->server));
		list_remove(SP_SERVICE_service_cache, entry);
		SP_FREE(entry);
	}
}

/**
 * @brief Reset the age of a cache entry.
 *
 * @param service_id ID of service whose cache entry's age is to be reset.
 */
void cache_reset_age(const char* service_id) {
	SP_SERVICE_service_cache_entry_t cmp_entry;
	strncpy(cmp_entry.service_id, service_id, SP_SERVICE_ID_LENGTH);
	SP_SERVICE_service_cache_entry_t* entry = (SP_SERVICE_service_cache_entry_t*) list_find(SP_SERVICE_service_cache, &cmp_entry, matches_entry_service_id);
	if(entry == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Trying to reset age of cache entry for service \"%s\", but not entry exists...\n", service_id);
		return;
	} else if(!entry->reachable) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Trying to reset age of cache entry for service \"%s\", but service is not reachable...\n", service_id);
		return;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SERVICE: Resetting age of cache entry for service \"%s\"...\n", service_id);
	entry->age = 0;
}

/**
 * @brief Check whether two cache entries have the same service ID and server
 * address.
 *
 * @param item Cache entry to check.
 * @param cmp_item Reference cache entry.
 * @return <code>true</code> if the two cache entries have the same service ID
 * and server address, <code>false</code> otherwise.
 */
static bool matches_entry_service_id_and_server(const void* item, const void* cmp_item) {
	SP_SERVICE_service_cache_entry_t* entry = (SP_SERVICE_service_cache_entry_t*) item;
	SP_SERVICE_service_cache_entry_t* cmp_entry = (SP_SERVICE_service_cache_entry_t*) cmp_item;
	return (
		same_service_id(entry->service_id, cmp_entry->service_id)
		&& (entry->server == cmp_entry->server )
	);
}

/**
 * @brief Callback that removes entries of unreachable services from the cache
 * after a certain time.
 *
 * @param parameter Cache entry that is to be removed.
 */
void cache_tag_not_reachable_callback(SP_callback_parameter_t parameter) {
	SP_SERVICE_service_cache_entry_t* callback_entry = (SP_SERVICE_service_cache_entry_t*) parameter;
	SP_SERVICE_service_cache_entry_t* entry = (SP_SERVICE_service_cache_entry_t*) list_find(SP_SERVICE_service_cache, callback_entry, matches_entry_service_id_and_server);
	if((entry != NULL) && (!entry->reachable)) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Removing entry for unreachable service \"%s\"...\n", entry->service_id);
		list_remove(SP_SERVICE_service_cache, entry);
		SP_FREE(entry);
	}
	SP_FREE(callback_entry);
}

/**
 * @brief Tag a service as not reachable.
 *
 * This is to avoid excessive queries for unreachable services. The tag is
 * removed after a certain time and query requests are processed again.
 *
 * @param service_id ID of service whose cache entry is to be tagged as not
 * reachable.
 */
void cache_tag_not_reachable(const char* service_id) {
	SP_SERVICE_service_cache_entry_t cmp_entry;
	strncpy(cmp_entry.service_id, service_id, SP_SERVICE_ID_LENGTH);
	SP_SERVICE_service_cache_entry_t* entry = (SP_SERVICE_service_cache_entry_t*) list_find(SP_SERVICE_service_cache, &cmp_entry, matches_entry_service_id);
	if((entry != NULL) && (!entry->reachable)) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SERVICE: Trying to tag cache entry for service \"%s\" not reachable, but entry is already tagged...\n", service_id);
		return;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SERVICE: Tagging cache entry for service \"%s\" not reachable and scheduling removal in %.2fs...\n", service_id, SP_TIMEOUT);
	if(entry == NULL) {
		SP_MALLOC(entry, SP_SERVICE_service_cache_entry_t);
		strncpy(entry->service_id, service_id, SP_SERVICE_ID_LENGTH);
		list_add(SP_SERVICE_service_cache, entry);
		entry->server = SP_BROADCAST_ADDR;
		entry->age = 0;
	}
	entry->reachable = false;
	SP_SERVICE_service_cache_entry_t* callback_entry;
	SP_MALLOC(callback_entry, SP_SERVICE_service_cache_entry_t);
	memcpy(callback_entry, entry, sizeof(SP_SERVICE_service_cache_entry_t));
	callback_add(cache_tag_not_reachable_callback, callback_entry, SP_TIMEOUT);
}
