/**
 * @file slp-light.c
 *
 * @brief Service Location Protocol (SLP) - minimal implemenation.
 *
 * This is a very basic implementation of the Service Location Protocol (SLP).
 * It covers searching for services by expanding-ring broadcasts. It does not
 * include service naming conventions or Directory Agents (DA).
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: slp-light.c 2565 2008-02-20 12:57:20Z wittenbu $
 */

#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "../types.h"
#include "../vars.h"
#include "../interface.h"
#include "../interface_internal.h"
#include "../util/debug.h"
#include "../util/list.h"
#include "../util/callback.h"
#include "../util/net.h"
#include "../rudp-light/rudp-light.h"
#include "../service/service.h"
#include "slp-light.h"
#include "slp-light_types.h"
#include "slp-light_vars.h"

#define SP_TYPE_SLP_QUERY		0x1	///< SLP QUERY packet type.
#define SP_TYPE_SLP_REPLY		0x2	///< SLP REPLY packet type.

/**
 * @brief Maximal supported number of active queries.
 */
#define SLP_MAX_ACTIVE_QUERIES		SP_MAX_SERVICES

/**
 * @brief Number of retries of SLP queries.
 *
 * With each retry of the query the TTL of the query packet is increased, thus
 * implementing an expanding ring search.
 */
#define SLP_QUERY_RETRIES		3

/**
 * @brief Timeout for SLP queries in seconds.
 */
#define SLP_QUERY_TIMEOUT		(SP_EXPECTED_MULTI_HOP_RTT + SP_ROUTE_ESTABLISHMENT_TIME)



/*
 * Interface to main module.
 */

/**
 * @brief Initialize the SLP module.
 */
void SLP_init() {
	list_alloc(&SP_SLP_active_queries);
	SP_SLP_total_queries = 0;
	SP_SLP_successful_queries = 0;
	SP_SLP_failed_queries = 0;
}

/**
 * @brief Start the SLP module.
 */
void SLP_start() {
}

/**
 * @brief Stop the SLP module.
 */
void SLP_stop() {
}

/**
 * @brief Finalize the SLP module.
 */
void SLP_finalize() {
	list_free(SP_SLP_active_queries);
	if(SP_SLP_total_queries != (SP_SLP_successful_queries + SP_SLP_failed_queries)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SLP: Query counters don't add up (%d != (%d + %d)).\n", SP_SLP_total_queries, SP_SLP_successful_queries, SP_SLP_failed_queries);
		abort();
	}
}



/*
 * Packet handling functions.
 */

/**
 * @brief Send a SLP QUERY packet.
 *
 * @param service_id ID of the service to query for.
 * @param retry Retry of this query.
 */
static void send_SLP_QUERY(const char* service_id, u_int8_t retry) {
	SP_packet_header_t header;
	SP_SLP_QUERY_packet_t query;

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SLP: Broadcasting QUERY for service \"%s\"...\n", service_id);
	init_header(&header);
	header.to = SP_BROADCAST_ADDR;
	header.ttl = (SP_ttl_t) pow(2, retry);
	if(header.ttl > SP_DEFAULT_TTL)
		header.ttl = SP_DEFAULT_TTL;
	header.app = SP_APP_SLP;
	header.type = SP_TYPE_SLP_QUERY;
	header.payload_length = sizeof(SP_SLP_QUERY_packet_t);
	strncpy(query.service_id, service_id, SP_SERVICE_ID_LENGTH);
	SP_send(&header, (SP_packet_payload_t*) &query);
}

/**
 * @brief Send a SLP REPLY packet.
 *
 * @param to Destination address of the reply.
 * @param service_id ID of the service that has been queried for.
 */
static void send_SLP_REPLY(SP_addr_t to, const char* service_id) {
	SP_packet_header_t header;
	SP_SLP_REPLY_packet_t reply;

	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SLP: Sending REPLY to query for service \"%s\" to node %s...\n", service_id, addr_to_str(to));
	init_header(&header);
	header.to = to;
	header.app = SP_APP_SLP;
	header.type = SP_TYPE_SLP_REPLY;
	header.payload_length = sizeof(SP_SLP_REPLY_packet_t);
	strncpy(reply.service_id, service_id, SP_SERVICE_ID_LENGTH);
	RUDP_send(&header, (SP_packet_payload_t*) &reply, NULL, NULL);
}

/**
 * @brief Handle an incoming SLP QUERY.
 *
 * Send a reply to this query if the service is hosted on this node, forward
 * the query otherwise.
 *
 * @param header Pointer to header of incoming packet.
 * @param query Pointer to SLP query.
 */
static void handle_SLP_QUERY(const SP_packet_header_t* header, const SP_SLP_QUERY_packet_t* query) {
	if(header->from == SP_LOCAL_ADDR) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SLP: Dropping own QUERY...\n");
		return;
	}
	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SLP: Received QUERY from node %s.\n", addr_to_str(header->from));
	if(SERVICE_get_active_service(query->service_id) != NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SLP: Service \"%s\" found.\n", query->service_id);
		send_SLP_REPLY(header->from, query->service_id);
	} else {
		// NOTE: One could also check the local service cache here.
		// However, this would get in the way of directory agents.
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SLP: Forwarding QUERY...\n");
		forward(header, (SP_packet_payload_t*) query);
	}
}

/**
 * @brief Check whether the service ID of two queries match.
 *
 * @param item Query to check.
 * @param cmp_item Reference query.
 * @return <code>true</code> if the service ID of the two queries matches,
 * <code>false</code> otherwise.
 */
static bool matches_query_service_id(const void* item, const void* cmp_item) {
	SP_SLP_query_t* query = (SP_SLP_query_t*) item;
	SP_SLP_query_t* cmp_query = (SP_SLP_query_t*) cmp_item;
	return SERVICE_same_service_id(query->service_id, cmp_query->service_id);
}

/**
 * @brief Check whether the server of two service instances match.
 *
 * @param item Service instance to check.
 * @param cmp_item Reference service instance.
 * @return <code>true</code> if the server of the two service instances
 * matches, <code>false</code> otherwise.
 */
static bool matches_service_instance_server(const void* item, const void* cmp_item) {
	SP_SLP_service_instance_t* service_instance = (SP_SLP_service_instance_t*) item;
	SP_SLP_service_instance_t* cmp_service_instance = (SP_SLP_service_instance_t*) cmp_item;
	return (service_instance->server == cmp_service_instance->server);
}

/**
 * @brief Handle an incoming SLP REPLY.
 *
 * Add or update server in the service instances list in the query results.
 *
 * @param header Pointer to the header of the incoming reply.
 * @param reply Pointer to the SLP reply.
 */
static void handle_SLP_REPLY(const SP_packet_header_t* header, const SP_SLP_REPLY_packet_t* reply) {
	SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SLP: Received REPLY to query for service \"%s\" from node %s.\n", reply->service_id, addr_to_str(header->from));

	SP_SLP_query_t cmp_query;
	strncpy(cmp_query.service_id, reply->service_id, SP_SERVICE_ID_LENGTH);
	SP_SLP_query_t* query = (SP_SLP_query_t*) list_find(SP_SLP_active_queries, &cmp_query, matches_query_service_id);
	if(query == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SLP: No matching query for service \"%s\" in active query list.\n", reply->service_id);
		return;
	}

	SP_SLP_service_instance_t cmp_service_instance;
	cmp_service_instance.server = header->from;
	SP_SLP_service_instance_t* service_instance = (SP_SLP_service_instance_t*) list_find(query->result->service_instances, &cmp_service_instance, matches_service_instance_server);
	if(service_instance == NULL) {
		SP_SLP_service_instance_t* service_instance;
		SP_MALLOC(service_instance, SP_SLP_service_instance_t);
		service_instance->server = header->from;
		service_instance->hops = header->hops;
		service_instance->delay = SP_time() - query->start_time;
		list_add(query->result->service_instances, service_instance);
		if(list_size(query->result->service_instances) >= query->min_service_instances) {
			SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SLP: Received enough replies to query for service \"%s\", calling callback function...\n", query->service_id);
			query->result->errno = SP_ERROR_OK;
			query->function(query->result);
			query->result = NULL;
			list_remove(SP_SLP_active_queries, query);
			// The query is freed in the timeout handler.
		}
	} else {
		if(service_instance->hops > header->hops)
			service_instance->hops = header->hops;
	}
}

/**
 * @brief SLP packet classifier.
 *
 * @param header Pointer to header of an incoming packet.
 * @param payload Pointer to payload of an incoming packet.
 */
void SLP_recv(const SP_packet_header_t* header, const SP_packet_payload_t* payload) {
	if(header->app != SP_APP_SLP) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SLP: Received non-SLP packet.\n");
		abort();
	}

	switch(header->type) {
	case SP_TYPE_NONE:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SLP: Received packet with uninitialized type field from node %s.\n", addr_to_str(header->from));
		break;
	case SP_TYPE_SLP_QUERY:
		if(header->payload_length != sizeof(SP_SLP_QUERY_packet_t)) {
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SLP: Received QUERY packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SP_SLP_QUERY_packet_t));
			abort();
		}
		handle_SLP_QUERY(header, (SP_SLP_QUERY_packet_t*) payload);
		break;
	case SP_TYPE_SLP_REPLY:
		if(header->payload_length != sizeof(SP_SLP_REPLY_packet_t)) {
			SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SLP: Received REPLY packet with wrong payload length (%d != %d).\n", header->payload_length, sizeof(SP_SLP_REPLY_packet_t));
			abort();
		}
		handle_SLP_REPLY(header, (SP_SLP_REPLY_packet_t*) payload);
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SLP: Received packet of unknown type 0x%X from node %s.\n", header->type, addr_to_str(header->from));
		abort();
	}
}



/*
 * Interface provided by this module.
 */

/**
 * @brief Timeout handler for locating a service.
 *
 * @param parameter Query for which the timeout has expired.
 */
static void SLP_locate_service_timeout(SP_callback_parameter_t parameter) {
	SP_SLP_query_t* query = (SP_SLP_query_t*) parameter;

	// Callback has already been called because enough instances were found.
	if(query->result == NULL) {
		SP_FREE(query);

		SP_SLP_successful_queries++;

	// Call callback function if any service instances have been found.
	} else if(!list_empty(query->result->service_instances)) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SLP: Timeout for QUERY for service \"%s\" has expired, calling callback function...\n", query->service_id);
		query->result->errno = SP_ERROR_OK;
		query->function(query->result);
		list_remove(SP_SLP_active_queries, query);
		SP_FREE(query);

		SP_SLP_successful_queries++;

	// Retry if no service instances have been found.
	} else if(query->retries > 0) {
		u_int8_t retry = SLP_QUERY_RETRIES - (--query->retries);
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SLP: Timeout for QUERY for service \"%s\" has expired (retry #%d).\n", query->service_id, retry);
		query->start_time = SP_time();	// Update start time for the next retry.
		send_SLP_QUERY(query->service_id, retry);
		callback_add(SLP_locate_service_timeout, query, SLP_QUERY_TIMEOUT);

	// Abort and call callback if no retries are left.
	} else {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SLP: Timeout for QUERY for service \"%s\" has expired. No retries left, service was not found.\n", query->service_id);
		query->result->errno = SP_ERROR_TIMEOUT;
		query->function(query->result);
		list_remove(SP_SLP_active_queries, query);
		SP_FREE(query);

		SP_SLP_failed_queries++;
	}
}

/**
 * @brief Locate a service.
 *
 * @param service_id Service ID of the service to be located.
 * @param min_service_instances Number of service instances found before the query is successful.
 * @param function Callback function.
 * @param result Pointer to query results.
 */
void SLP_locate_service(const char* service_id, unsigned int min_service_instances, SP_callback_function_t function, SP_SLP_query_result_t* result) {
	if((strlen(service_id) + 1) > SP_SERVICE_ID_LENGTH) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SLP: Service ID \"%s\" is too long (%d + 1 > %d).\n", service_id, strlen(service_id), SP_SERVICE_ID_LENGTH);
		result->errno = SP_ERROR_PARAMETER;
		function(result);
		return;
	} else if((function == NULL) || (result == NULL)) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SLP: Callback function or parameter not given (%p, %p).\n", function, result);
		result->errno = SP_ERROR_PARAMETER;
		function(result);
		return;
	} else if(!list_empty(result->service_instances)) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SLP: Query result list for service instances is not empty.\n");
		result->errno = SP_ERROR_PARAMETER;
		function(result);
		return;
	}

	SP_SLP_total_queries++;

	if(list_size(SP_SLP_active_queries) >= SLP_MAX_ACTIVE_QUERIES) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SLP: No empty slots for query processing.\n");
		result->errno = SP_ERROR_BUSY;
		function(result);
		SP_SLP_failed_queries++;
		return;
	} else if(SERVICE_get_active_service(service_id) != NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SLP: Service \"%s\" is hosted on the node itself.\n", service_id);
		SP_SLP_service_instance_t* service_instance;
		SP_MALLOC(service_instance, SP_SLP_service_instance_t);
		service_instance->server = SP_LOCAL_ADDR;
		service_instance->hops = 0;
		service_instance->delay = 0.0;
		list_add(result->service_instances, service_instance);
		result->errno = SP_ERROR_OK;
		function(result);
		SP_SLP_successful_queries++;
		return;
	}

	SP_SLP_query_t* query;
	SP_MALLOC(query, SP_SLP_query_t);
	strncpy(query->service_id, service_id, SP_SERVICE_ID_LENGTH);
	query->min_service_instances = min_service_instances;
	query->retries = SLP_QUERY_RETRIES;
	query->start_time = SP_time();
	query->function = function;
	query->result = result;
	strncpy(query->result->service_id, service_id, SP_SERVICE_ID_LENGTH);
	query->result->errno = SP_ERROR_NOT_SET;

	if(list_find(SP_SLP_active_queries, query, matches_query_service_id) != NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "SLP: Query for service \"%s\" already in progress.\n", service_id);
		result->errno = SP_ERROR_IN_PROGRESS;
		function(result);
		SP_FREE(query);
		SP_SLP_total_queries--;
		return;
	}

	SP_DEBUG(SP_DEBUG_LEVEL_INFO, "SLP: Locating service \"%s\"...\n", service_id);
	list_add(SP_SLP_active_queries, query);
	send_SLP_QUERY(service_id, 0);
	callback_add(SLP_locate_service_timeout, query, SLP_QUERY_TIMEOUT);
}



/*
 * Interface provided by this module for direct commands.
 */

/**
 * @brief Callback of command function to locate a service.
 *
 * @param parameter Result of the query.
 */
static void SLP_command_locate_service_callback(SP_callback_parameter_t parameter) {
	SP_SLP_query_result_t* result = (SP_SLP_query_result_t*) parameter;

	SP_SLP_service_instance_t* service_instance;
	switch(result->errno) {
	case SP_ERROR_OK:
		service_instance = (SP_SLP_service_instance_t*) list_head(result->service_instances);
		SP_DEBUG(SP_DEBUG_LEVEL_NOTE, "SLP: %d instance(s) of service \"%s\" located on node(s) %s (%d hops)", list_size(result->service_instances), result->service_id, addr_to_str(service_instance->server), service_instance->hops);
		list_remove(result->service_instances, service_instance);
		while(!list_empty(result->service_instances)) {
			service_instance = (SP_SLP_service_instance_t*) list_head(result->service_instances);
			list_remove(result->service_instances, service_instance);
			if(list_size(result->service_instances) >= 1) {
				SP_DEBUG(SP_DEBUG_LEVEL_NONE, ", ");
			} else {
				SP_DEBUG(SP_DEBUG_LEVEL_NONE, " and ");
			}
			SP_DEBUG(SP_DEBUG_LEVEL_NONE, "%s (%d hops)", addr_to_str(service_instance->server), service_instance->hops);
		}
		SP_DEBUG(SP_DEBUG_LEVEL_NONE, ".\n");
		break;
	case SP_ERROR_PARAMETER:
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "SLP: Wrong parameter value.\n");
		break;
	case SP_ERROR_TIMEOUT:
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SLP: Service \"%s\" not found.\n", result->service_id);
		break;
	case SP_ERROR_BUSY:
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SLP: Too many service queries in progress.\n");
		break;
	case SP_ERROR_IN_PROGRESS:
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "SLP: Query for service \"%s\" already in progress.\n", result->service_id);
		break;
	default:
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "SLP: Callback with unknown error code 0x%X.\n", result->errno);
		abort();
	}

	list_free(result->service_instances);
	SP_FREE(result);
}

/**
 * @brief SLP command function.
 *
 * @param argc Number of command parameters.
 * @param argv Array of strings used as command parameters.
 * @return <code>true</code> if the command was executed successfully,
 * <code>false</code> otherwise.
 */
bool SLP_command(int argc, const char*const* argv) {
	if(argc == 4) {
		if(strcmp(argv[1], "SLP_locate_service") == 0) {
			SP_SLP_query_result_t* result;
			SP_MALLOC(result, SP_SLP_query_result_t);
			list_alloc(&result->service_instances);
			SLP_locate_service(argv[2], atoi(argv[3]), SLP_command_locate_service_callback, result);
			return true;
		}
	}
	return false;
}
