/**
 * @file callback.c
 *
 * @brief Callback functions.
 *
 * Copyright 2007-2008, Freie Universitaet Berlin. All rights reserved.
 * Please refer to the COPYING file for detailed licensing information.
 *
 * @author Georg Wittenburg
 *
 * $Id: callback.c 2570 2008-02-20 17:26:26Z wittenbu $
 */

#include "../types.h"
#include "../vars.h"
#include "../interface.h"
#include "../interface_internal.h"
#include "debug.h"
#include "list.h"
#include "callback.h"

/**
 * @brief Maximal supported number of simultaneous callbacks.
 */
#define MAX_CALLBACKS	1024

/**
 * @brief Check whether a callback expires later than another one.
 *
 * @param item Pointer to callback to check.
 * @param cmp_item Pointer to reference callback
 * @return <code>true</code> if the callback to check expires later than the
 * reference callback, <code>false</code> otherwise.
 */
static bool is_later(const void* item, const void* cmp_item) {
	SP_callback_t* callback = (SP_callback_t*) item;
	SP_callback_t* cmp_callback = (SP_callback_t*) cmp_item;
	return (callback->time > cmp_callback->time);
}

/**
 * @brief Add a new callback to be executed later on.
 *
 * @param function Pointer to the callback function.
 * @param parameter Pointer to the parameter passed to the callback function.
 * @param delay Delay before the callback expires in seconds.
 */
void callback_add(SP_callback_function_t function, SP_callback_parameter_t parameter, SP_time_t delay) {
	if(function == NULL) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Trying to register callback with NULL function...\n");
		return;
	}
	if(delay < 0) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "Trying to register callback to function at %p with negative delay (%f)...\n", delay);
		return;
	}
	if(list_size(SP_callbacks) >= MAX_CALLBACKS) {
		SP_DEBUG(SP_DEBUG_LEVEL_ERROR, "Trying to register callback to function at %p, but %d callbacks are already running...\n", function, MAX_CALLBACKS);
		return;
	}

	if(delay == 0) {
		SP_DEBUG(SP_DEBUG_LEVEL_DEBUG, "Registering callback to function at %p and parameter at %p without delay...\n", function, parameter);
	} else {
		SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Registering callback to function at %p and parameter at %p with a delay of %.2f seconds...\n", function, parameter, delay);
	}
	SP_callback_t* callback;
	SP_MALLOC(callback, SP_callback_t);
	callback->function = function;
	callback->parameter = parameter;
	callback->time = SP_time() + delay;
	list_insert_where(SP_callbacks, callback, is_later);
	SP_activate_callback(((SP_callback_t*) list_head(SP_callbacks))->time);
	SP_registered_callbacks++;
}

/**
 * @brief This function is called when a callback has expired.
 */
void callback_expire() {
	if(list_empty(SP_callbacks)) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Trying to call expired callback, but no callbacks have been registered.\n");
		abort();
	} else if(SP_callback_time == 0) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Trying to call expired callback, but time for next callback has not been set.\n");
		abort();
	}
	SP_time_t time = SP_time();
	SP_callback_t* callback = (SP_callback_t*) list_head(SP_callbacks);
	if((callback->time - time) > 0) {
		SP_DEBUG(SP_DEBUG_LEVEL_FATAL, "Trying to call expired callback, but next callback is only due in %.2f seconds.\n", callback->time - time);
		abort();
	}

	SP_DEBUG(SP_DEBUG_LEVEL_VERBOSE, "Calling expired callback function at %p with parameter at %p (%f seconds overdue)...\n", callback->function, callback->parameter, time - callback->time);
	if((callback->time - time) <= -1) {
		SP_DEBUG(SP_DEBUG_LEVEL_WARN, "Callback is overdue by %f seconds.\n", callback->time - time);
	}
	SP_callback_time = 0;
	list_remove(SP_callbacks, callback);
	callback->function(callback->parameter);
	SP_FREE(callback);
	SP_executed_callbacks++;

	// Schedule next callback.
	if(!list_empty(SP_callbacks))
		SP_activate_callback(((SP_callback_t*) list_head(SP_callbacks))->time);
}
